#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def object_reference(name, *args, **kwargs):
    "Macro to cut down on a lot of code"
    block_name = kwargs.pop('block_name', name + 's').replace(' ', '_')

    dyn_type_path = ".....%s_palette.STEPTREE[DYN_I].name.filepath" % block_name
    return Struct(name,
        dyn_senum16('type', DYN_NAME_PATH=dyn_type_path),
        dyn_senum16('name',
            DYN_NAME_PATH=".....object_names.STEPTREE[DYN_I].name"),
        Bool16('not_placed',
            "automatically",
            "on_easy",
            "on_normal",
            "on_hard",
            ),
        SInt16('desired_permutation'),
        QStruct("position", INCLUDE=xyz_float),
        ypr_float_rad("rotation"),
        *args,
        **kwargs
        )

def object_swatch(name, def_id, size=48):
    "Macro to cut down on a lot of code"
    return Struct(name,
        dependency("name", def_id),
        SIZE=size
        )

fl_float_xyz = QStruct("",
    FlFloat("x"),
    FlFloat("y"),
    FlFloat("z"),
    ORIENT="h"
    )

stance_flags = FlBool16("stance",
    "walk",
    "look_only",
    "primary_fire",
    "secondary_fire",
    "jump",
    "crouch",
    "melee",
    "flashlight",
    "action1",
    "action2",
    "action_hold",
    )

unit_control_packet = Struct("unit_control_packet",

    )

r_a_stream_header = Struct("r_a_stream_header",
    UInt8("move_index", DEFAULT=3, MAX=6),
    UInt8("bool_index"),
    stance_flags,
    FlSInt16("weapon", DEFAULT=-1),
    QStruct("speed", FlFloat("x"), FlFloat("y"), ORIENT="h"),
    QStruct("feet", INCLUDE=fl_float_xyz),
    QStruct("body", INCLUDE=fl_float_xyz),
    QStruct("head", INCLUDE=fl_float_xyz),
    QStruct("change", INCLUDE=fl_float_xyz),
    FlUInt16("unknown1"),
    FlUInt16("unknown2"),
    FlUInt16("unknown3", DEFAULT=0xFFFF),
    FlUInt16("unknown4", DEFAULT=0xFFFF),
    SIZE=60
    )

device_flags = (
    "initially_open",  # value of 1.0
    "initially_off",  # value of 0.0
    "can_change_only_once",
    "position_reversed",
    "not_usable_from_any_side"
    )

location_types = (
    "none",
    "ctf",
    "slayer",
    "oddball",
    "king",
    "race",
    "terminator",
    "stub",
    "ignored1",
    "ignored2",
    "ignored3",
    "ignored4",
    "all_games",
    "all_games_except_ctf",
    "all_games_except_ctf_and_race"
    )

group_indices = tuple("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

squad_states = (
    "none",
    "sleeping",
    "alert",
    "moving_repeat_same_position",
    "moving_loop",
    "moving_loop_back_and_forth",
    "moving_loop_randomly",
    "moving_randomly",
    "guarding",
    "guarding_at_position",
    "searching",
    "fleeing"
    )

maneuver_when_states = (
    "never",
    {NAME: "strength_at_75_percent", GUI_NAME: "< 75% strength"},
    {NAME: "strength_at_50_percent", GUI_NAME: "< 50% strength"},
    {NAME: "strength_at_25_percent", GUI_NAME: "< 25% strength"},
    "anybody_dead",
    {NAME: "dead_at_25_percent", GUI_NAME: "25% dead"},
    {NAME: "dead_at_50_percent", GUI_NAME: "50% dead"},
    {NAME: "dead_at_75_percent", GUI_NAME: "75% dead"},
    "all_but_one_dead",
    "all_dead"
    )

atom_types = (
    "pause",
    "goto",
    "goto_and_face",
    "move_in_direction",
    "look",
    "animation_mode",
    "crouch",
    "shoot",
    "grenade",
    "vehicle",
    "running_jump",
    "targeted_jump",
    "script",
    "animate",
    "recording",
    "action",
    "vocalize",
    "targeting",
    "initiative",
    "wait",
    "loop",
    "die",
    "move_immediate",
    "look_random",
    "look_player",
    "look_object",
    "set_radius",
    "teleport"
    )


sky = Struct("sky",
    dependency("sky", "sky "),
    SIZE=16
    )

child_scenario = Struct("child_scenario",
    dependency("child_scenario", "scnr"),
    SIZE=32
    )

function = Struct('function',
    Bool32('flags',
        'scripted',
        'invert',
        'additive',
        'always_active',
        ),
    ascii_str32('name'),
    float_sec('period'),  # seconds
    dyn_senum16('scale_period_by', DYN_NAME_PATH="..[DYN_I].name"),
    SEnum16('function', *animation_functions),
    dyn_senum16('scale_function_by', DYN_NAME_PATH="..[DYN_I].name"),
    SEnum16('wobble_function', *animation_functions),
    float_sec('wobble_period'),  # seconds
    Float('wobble_magnitude', SIDETIP="%"),  # percent
    Float('square_wave_threshold'),
    SInt16('step_count'),
    SEnum16('map_to', *fade_functions),
    SInt16('sawtooth_count'),

    Pad(2),
    dyn_senum16('scale_result_by', DYN_NAME_PATH="..[DYN_I].name"),
    SEnum16('bounds_mode',
        'clip',
        'clip_and_normalize',
        'scale_to_fit',
        ),
    QStruct('bounds', INCLUDE=from_to),

    Pad(6),
    dyn_senum16('turn_off_with', DYN_NAME_PATH="..[DYN_I].name"),

    SIZE=120
    )

comment = Struct("comment",
    QStruct("position", INCLUDE=xyz_float),
    Pad(16),
    rawtext_ref("comment_data", StrLatin1, max_size=16384),
    SIZE=48
    )

object_name = Struct("object_name",
    ascii_str32("name"),
    FlSEnum16("object_type",
        *((object_types[i], i - 1) for i in
          range(len(object_types))),
        VISIBLE=False, EDITABLE=False, DEFAULT=-1
        ),
    FlSInt16("reflexive_index", VISIBLE=False, EDITABLE=False),
    SIZE=36
    )

# Object references
scenery = object_reference("scenery", SIZE=72, block_name="sceneries")

biped = object_reference("biped",
    Pad(40),
    float_zero_to_one("body_vitality"),
    Bool32("flags",
        "dead",
        ),
    SIZE=120
    )

vehicle = object_reference("vehicle",
    Pad(40),
    float_zero_to_one("body_vitality"),
    Bool32("flags",
        "dead",
        ),

    Pad(8),
    SInt8("multiplayer_team_index"),
    Pad(1),
    Bool16("multiplayer_spawn_flags",
        "slayer_default",
        "ctf_default",
        "king_default",
        "oddball_default",
        #"unused1",
        #"unused2",
        #"unused3",
        #"unused4",
        ("slayer_allowed", 1<<8),
        ("ctf_allowed", 1<<9),
        ("king_allowed", 1<<10),
        ("oddball_allowed", 1<<11),
        #"unused5",
        #"unused6",
        #"unused7",
        #"unused8",
        ),
    SIZE=120
    )

equipment = object_reference("equipment",
    Pad(2),
    Bool16("flags",
        "initially_at_rest",
        "obsolete",
        {NAME: "can_accelerate", GUI_NAME:"moves due to explosions"},
        ),
    SIZE=40
    )

weapon = object_reference("weapon",
    Pad(40),
    SInt16("rounds_left"),
    SInt16("rounds_loaded"),
    Bool16("flags",
        "initially_at_rest",
        "obsolete",
        {NAME: "can_accelerate", GUI_NAME:"moves due to explosions"},
        ),
    SIZE=92
    )

device_group = Struct("device_group",
    ascii_str32("name"),
    float_zero_to_one("initial_value"),
    Bool32("flags",
        "can_change_only_once"
        ),
    SIZE=52
    )

machine = object_reference("machine",
    Pad(8),
    dyn_senum16("power_group",
        DYN_NAME_PATH=".....device_groups.STEPTREE[DYN_I].name"),
    dyn_senum16("position_group",
        DYN_NAME_PATH=".....device_groups.STEPTREE[DYN_I].name"),
    Bool32("flags", *device_flags),
    Bool32("more_flags",
        "does_not_operate_automatically",
        "one_sided",
        "never_appears_locked",
        "opened_by_melee_attack",
        ),
    SIZE=64
    )

control = object_reference("control",
    Pad(8),
    dyn_senum16("power_group",
        DYN_NAME_PATH=".....device_groups.STEPTREE[DYN_I].name"),
    dyn_senum16("position_group",
        DYN_NAME_PATH=".....device_groups.STEPTREE[DYN_I].name"),
    Bool32("flags", *device_flags),
    Bool32("more_flags",
        "usable_from_both_sides",
        ),
    SInt16("DONT_TOUCH_THIS"),  # why?
    SIZE=64
    )

light_fixture = object_reference("light_fixture",
    Pad(8),
    dyn_senum16("power_group",
        DYN_NAME_PATH=".....device_groups.STEPTREE[DYN_I].name"),
    dyn_senum16("position_group",
        DYN_NAME_PATH=".....device_groups.STEPTREE[DYN_I].name"),
    Bool32("flags", *device_flags),
    QStruct("color", INCLUDE=rgb_float),
    Float("intensity"),
    Float("falloff_angle"),  # radians
    Float("cutoff_angle"),  # radians
    SIZE=88
    )

sound_scenery = object_reference("sound_scenery", SIZE=40, block_name="sound_sceneries")

# Object swatches
scenery_swatch = object_swatch("scenery_swatch", "scen")
biped_swatch = object_swatch("biped_swatch", "bipd")
vehicle_swatch = object_swatch("vehicle_swatch", "vehi")
equipment_swatch = object_swatch("equipment_swatch", "eqip")
weapon_swatch = object_swatch("weapon_swatch", "weap")
machine_swatch = object_swatch("machine_swatch", "mach")
control_swatch = object_swatch("control_swatch", "ctrl")
light_fixture_swatch = object_swatch("light_fixture_swatch", "lifi")
sound_scenery_swatch = object_swatch("sound_scenery_swatch", "ssce")

player_starting_profile = Struct("player_starting_profile",
    ascii_str32("name"),
    float_zero_to_one("starting_health_modifier"),
    float_zero_to_one("starting_shield_modifier"),
    dependency("primary_weapon", "weap"),
    SInt16("primary_rounds_loaded"),
    SInt16("primary_rounds_total"),
    dependency("secondary_weapon", "weap"),
    SInt16("secondary_rounds_loaded"),
    SInt16("secondary_rounds_total"),
    SInt8("starting_frag_grenade_count", MIN=0),
    SInt8("starting_plasma_grenade_count", MIN=0),
    SIZE=104
    )

player_starting_location = Struct("player_starting_location",
    QStruct("position", INCLUDE=xyz_float),
    float_rad("facing"),  # radians
    SInt16("team_index"),
    dyn_senum16("bsp_index",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath"),
    SEnum16("type_0", *location_types),
    SEnum16("type_1", *location_types),
    SEnum16("type_2", *location_types),
    SEnum16("type_3", *location_types),
    SIZE=52
    )

player_starting_location2 = dict(player_starting_location)
player_starting_location2[3] = dyn_senum16("bsp_index",
        DYN_NAME_PATH=("........structure_bsps.STEPTREE"
                       "[DYN_I].structure_bsp.filepath")
        )

trigger_volume = Struct("trigger_volume",
    # if this unknown != 1, the trigger volume is disabled
    FlUInt16("unknown0", DEFAULT=1, EDITABLE=False, VISIBLE=False),
    Pad(2),
    ascii_str32("name"),
    BytesRaw("unknown1", SIZE=12, VISIBLE=False),
    QStruct("binormal", INCLUDE=ijk_float),
    QStruct("normal",   INCLUDE=ijk_float),
    QStruct("position", INCLUDE=xyz_float,
        TOOLTIP=("Volume sides extend in one direction from this position.\n"
                 "This position is the origin corner for the trigger volume.")),
    QStruct("sides",
        Float("w", TOOLTIP="Along local y axis"),
        Float("h", TOOLTIP="Along local z axis"),
        Float("l", TOOLTIP="Along local x axis"),
        ORIENT='h'
        ),
    SIZE=96,
    COMMENT=(
        "To make the trigger volumes rotation be zero, the normal and\n"
        "binormal should be (1, 0, 0) and (0, 1, 0) respectively.")
    )

recorded_animation = Struct("recorded_animation",
    ascii_str32("name"),
    SInt8("version"),
    SInt8("raw_animation_data"),
    SInt8("unit_control_data_version"),
    Pad(1),
    SInt16("length_of_animation", SIDETIP="ticks"),  # ticks
    Pad(6),
    rawdata_ref("recorded_animation_event_stream", max_size=2097152),
    SIZE=64
    )

netgame_flag = Struct("netgame_flag",
    QStruct("position", INCLUDE=xyz_float),
    float_rad("facing"),  # radians
    SEnum16("type",
        "ctf_flag",
        "ctf_vehicle",
        "oddball_ball_spawn",
        "race_track",
        "race_vehicle",
        "vegas_bank",
        "teleport_from",
        "teleport_to",
        "hill_flag",
        ),
    SInt16("team_index"),
    dependency("weapon_group", "itmc"),
    SIZE=148
    )

netgame_equipment = Struct("netgame_equipment",
    Bool32("flags",
        "levitate"
        ),
    SEnum16("type_0", *location_types),
    SEnum16("type_1", *location_types),
    SEnum16("type_2", *location_types),
    SEnum16("type_3", *location_types),
    SInt16("team_index"),
    SInt16("spawn_time", SIDETIP="seconds(0 = default)",
            UNIT_SCALE=sec_unit_scale),  # seconds

    Pad(48),
    QStruct("position", INCLUDE=xyz_float),
    float_rad("facing"),  # radians
    dependency("item_collection", "itmc"),
    SIZE=144
    )

starting_equipment = Struct("starting_equipment",
    Bool32("flags",
        "no_grenades",
        "plasma_grenades",
        ),
    SEnum16("type_0", *location_types),
    SEnum16("type_1", *location_types),
    SEnum16("type_2", *location_types),
    SEnum16("type_3", *location_types),

    Pad(48),
    dependency("item_collection_1", "itmc"),
    dependency("item_collection_2", "itmc"),
    dependency("item_collection_3", "itmc"),
    dependency("item_collection_4", "itmc"),
    dependency("item_collection_5", "itmc"),
    dependency("item_collection_6", "itmc"),
    SIZE=204
    )

bsp_switch_trigger_volume = Struct("bsp_switch_trigger_volume",
    dyn_senum16("trigger_volume",
        DYN_NAME_PATH=".....trigger_volumes.STEPTREE[DYN_I].name"),
    dyn_senum16("source",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath"),
    dyn_senum16("destination",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath"),
    FlUInt16("unknown", EDITABLE=False),
    SIZE=8
    )

decal = Struct("decal",
    dyn_senum16("decal_type",
        DYN_NAME_PATH=".....decals_palette.STEPTREE[DYN_I].name.filepath"),
    SInt8("yaw"),
    SInt8("pitch"),
    QStruct("position", INCLUDE=xyz_float),
    SIZE=16
    )

decal_swatch = object_swatch("decal_swatch", "deca", 16)
detail_object_collection_swatch = object_swatch(
    "detail_object_collection_swatch", "dobc")
actor_swatch = object_swatch("actor_swatch", "actv", 16)

ai_anim_reference = Struct("ai_animation_reference",
    ascii_str32("animation_name"),
    dependency("animation_graph", "antr"),
    SIZE=60
    )

ai_script_reference = Struct("ai_script_reference",
    ascii_str32("script_name"),
    SIZE=40
    )

ai_recording_reference = Struct("ai_recording_reference",
    ascii_str32("recording_name"),
    SIZE=40
    )

halo_script = Struct("script",
    ascii_str32("name", EDITABLE=False),
    SEnum16("type", *script_types),
    SEnum16("return_type", *script_object_types, EDITABLE=False),
    UInt32("root_expression_index", EDITABLE=False),
    Computed("decompiled_script", WIDGET=HaloScriptTextFrame),
    SIZE=92,
    )

halo_global = Struct("global",
    ascii_str32("name", EDITABLE=False),
    SEnum16("type", *script_object_types, EDITABLE=False),
    Pad(6),
    UInt32("initialization_expression_index", EDITABLE=False),
    Computed("decompiled_script", WIDGET=HaloScriptTextFrame),
    SIZE=92,
    )

reference = Struct("tag_reference",
    Pad(24),
    dependency("reference"),
    SIZE=40
    )

source_file = Struct("source_file",
    ascii_str32("source_name"),
    rawdata_ref("source", max_size=262144, widget=HaloScriptSourceFrame),
    SIZE=52
    )

cutscene_flag = Struct("cutscene_flag",
    Pad(4),
    ascii_str32("name"),
    QStruct("position", INCLUDE=xyz_float),
    yp_float_rad("facing"),  # radians
    SIZE=92
    )

cutscene_camera_point = Struct("cutscene_camera_point",
    Pad(4),
    ascii_str32("name"),
    Pad(4),
    QStruct("position", INCLUDE=xyz_float),
    ypr_float_rad("orientation"),  # radians
    float_rad("field_of_view"),  # radians
    SIZE=104
    )

cutscene_title = Struct("cutscene_title",
    Pad(4),
    ascii_str32("name"),
    Pad(4),
    QStruct("text_bounds",
        SInt16("t"), SInt16("l"), SInt16("b"), SInt16("r"),
        ORIENT='h',
        ),
    SInt16("string_index"),
    SEnum16("text_style",
        "plain",
        "bold",
        "italic",
        "condense",
        "underline",
        ),
    SEnum16("justification",
        "left",
        "right",
        "center",
        ),

    Pad(6),
    #QStruct("text_color", INCLUDE=argb_byte),
    #QStruct("shadow_color", INCLUDE=argb_byte),
    UInt32("text_color", INCLUDE=argb_uint32),
    UInt32("shadow_color", INCLUDE=argb_uint32),
    float_sec("fade_in_time"),  # seconds
    float_sec("up_time"),  # seconds
    float_sec("fade_out_time"),  # seconds
    SIZE=96
    )


move_position = Struct("move_position",
    QStruct("position", INCLUDE=xyz_float),
    float_rad("facing"),  # radians
    Float("weight"),
    from_to_sec("time"),
    dyn_senum16("animation",
        DYN_NAME_PATH="tagdata.ai_animation_references.STEPTREE[DYN_I].animation_name"),
    SInt8("sequence_id"),

    Pad(45),
    SInt32("surface_index"),
    SIZE=80
    )

actor_starting_location = Struct("starting_location",
    QStruct("position", INCLUDE=xyz_float),
    float_rad("facing"),  # radians
    Pad(2),
    SInt8("sequence_id"),
    Bool8("flags",
        "required",
        ),
    SEnum16("return_state",  *(dict(NAME=n, GUI_NAME=n) for n in squad_states)),
    SEnum16("initial_state", *(dict(NAME=n, GUI_NAME=n) for n in squad_states)),
    dyn_senum16("actor_type",
        DYN_NAME_PATH="tagdata.actors_palette.STEPTREE[DYN_I].name.filepath"),
    dyn_senum16("command_list",
        DYN_NAME_PATH="tagdata.command_lists.STEPTREE[DYN_I].name"),
    SIZE=28
    )

squad = Struct("squad",
    ascii_str32("name"),
    dyn_senum16("actor_type",
        DYN_NAME_PATH="tagdata.actors_palette.STEPTREE[DYN_I].name.filepath"),
    dyn_senum16("platoon",
        DYN_NAME_PATH=".....platoons.STEPTREE[DYN_I].name"),
    SEnum16("initial_state", *(dict(NAME=n, GUI_NAME=n) for n in squad_states)),
    SEnum16("return_state",  *(dict(NAME=n, GUI_NAME=n) for n in squad_states)),
    Bool32("flags",
        "unused",
        "never_search",
        "start_timer_immediately",
        "no_timer_delay_forever",
        "magic_sight_after_timer",
        "automatic_migration",
        ),
    SEnum16("unique_leader_type",
        "normal",
        "none",
        "random",
        "sgt_johnson",
        "sgt_lehto",
        ),

    Pad(32),
    dyn_senum16("maneuver_to_squad", DYN_NAME_PATH="..[DYN_I].name"),
    float_sec("squad_delay_time"),  # seconds
    Bool32("attacking", *group_indices),
    Bool32("attacking_search", *group_indices),
    Bool32("attacking_guard", *group_indices),
    Bool32("defending", *group_indices),
    Bool32("defending_search", *group_indices),
    Bool32("defending_guard", *group_indices),
    Bool32("pursuing", *group_indices),

    Pad(12),
    SInt16("normal_diff_count"),
    SInt16("insane_diff_count"),
    SEnum16("major_upgrade",
        "normal",
        "few",
        "many",
        "none",
        "all",
        ),
    Pad(2),

    SInt16("respawn_min_actors"),
    SInt16("respawn_max_actors"),
    SInt16("respawn_total"),
    Pad(2),

    from_to_sec("respawn_delay"),

    Pad(48),
    reflexive("move_positions", move_position, 31),
    reflexive("starting_locations", actor_starting_location, 31),
    SIZE=232
    )

platoon = Struct("platoon",
    ascii_str32("name"),
    Bool32("flags",
        "flee_when_maneuvering",
        "say_advancing_when_maneuvering",
        "start_in_defending_state",
        ),

    Pad(12),
    SEnum16("change_attacking_defending_state", *maneuver_when_states),
    dyn_senum16("change_happens_to", DYN_NAME_PATH="..[DYN_I].name"),

    Pad(8),
    SEnum16("maneuver_when", *maneuver_when_states),
    dyn_senum16("maneuver_happens_to", DYN_NAME_PATH="..[DYN_I].name"),
    SIZE=172
    )

firing_position = Struct("firing_position",
    QStruct("position", INCLUDE=xyz_float),
    SEnum16("group_index", *group_indices),
    FlUInt16('bsp_cluster', VISIBLE=False),  # calculated on map compile
    Pad(4),
    FlSInt32('bsp_surface', VISIBLE=False),  # calculated on map compile
    SIZE=24
    )

encounter = Struct("encounter",
    ascii_str32("name"),
    Bool32("flags",
        "not_initially_created",
        "respawn_enabled",
        "initially_blind",
        "initially_deaf",
        "initially_braindead",
        "firing_positions_are_3d",
        "manual_bsp_index_specified",
        ),
    SEnum16("team_index",
        {NAME: "default",  GUI_NAME: "0 / default_by_unit"},
        {NAME: "player",   GUI_NAME: "1 / player"},
        {NAME: "human",    GUI_NAME: "2 / human"},
        {NAME: "covenant", GUI_NAME: "3 / covenant"},
        {NAME: "flood",    GUI_NAME: "4 / flood"},
        {NAME: "sentinel", GUI_NAME: "5 / sentinel"},
        {NAME: "unused6",  GUI_NAME: "6 / unused6"},
        {NAME: "unused7",  GUI_NAME: "7 / unused7"},
        {NAME: "unused8",  GUI_NAME: "8 / unused8"},
        {NAME: "unused9",  GUI_NAME: "9 / unused9"}
        ),
    SInt16('unknown', VISIBLE=False),
    SEnum16("search_behavior",
        "normal",
        "never",
        "tenacious"
        ),
    dyn_senum16("manual_bsp_index",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath"),
    from_to_sec("respawn_delay"),

    Pad(74),
    FlSInt16('bsp_index', VISIBLE=False),  # calculated on map compile
    reflexive("squads", squad, 64),
    reflexive("platoons", platoon, 32, DYN_NAME_PATH='.name'),
    reflexive("firing_positions", firing_position, 512),
    reflexive("player_starting_locations", player_starting_location2, 256),

    SIZE=176
    )

command = Struct("command",
    SEnum16("atom_type", *atom_types),
    SInt16("atom_modifier"),
    Float("parameter_1"),
    Float("parameter_2"),
    dyn_senum16("point_1", DYN_NAME_PATH=".....points.STEPTREE[DYN_I]"),
    dyn_senum16("point_2", DYN_NAME_PATH=".....points.STEPTREE[DYN_I]"),
    dyn_senum16("animation",
        DYN_NAME_PATH="tagdata.ai_animation_references.STEPTREE[DYN_I].animation_name"),
    dyn_senum16("script",
        DYN_NAME_PATH="tagdata.scripts.STEPTREE[DYN_I].name"),
    dyn_senum16("recording",
        DYN_NAME_PATH="tagdata.ai_recording_references.STEPTREE[DYN_I].recording_name"),
    dyn_senum16("command", DYN_NAME_PATH="..[DYN_I].atom_type.enum_name"),
    dyn_senum16("object_name", DYN_NAME_PATH="tagdata.object_names.STEPTREE[DYN_I].name"),
    SIZE=32
    )

point = Struct("point",
    QStruct("position", INCLUDE=xyz_float),
    SIZE=20
    )

command_list = Struct("command_list",
    ascii_str32("name"),
    Bool32("flags",
        "allow_initiative",
        "allow_targeting",
        "disable_looking",
        "disable_communication",
        "disable_falling_damage",
        "manual_bsp_index",
        ),

    Pad(8),
    dyn_senum16("manual_bsp_index",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath"),
    dyn_senum16("bsp_index",
        DYN_NAME_PATH=".....structure_bsps.STEPTREE[DYN_I].structure_bsp.filepath",
        VISIBLE=False
        ),
    reflexive("commands", command, 64),
    reflexive("points", point, 64),
    SIZE=96
    )

participant = Struct("participant",
    Pad(2),
    Bool16("flags",
        "optional",
        "has_alternate",
        "is_alternate",
        ),
    SEnum16("selection_type",
        "friendly_actor",
        "disembodied",
        "in_players_vehicle",
        "not_in_a_vehicle",
        "prefer_sergeant",
        "any_actor",
        "radio_unit",
        "radio_sergeant",
        ),
    SEnum16("actor_type", *actor_types),
    dyn_senum16("use_this_object", DYN_NAME_PATH="tagdata.object_names.STEPTREE[DYN_I].name"),
    dyn_senum16("set_new_name", DYN_NAME_PATH="tagdata.object_names.STEPTREE[DYN_I].name"),

    Pad(12),
    BytesRaw("unknown", DEFAULT=b"\xFF"*12, SIZE=12, VISIBLE=False),
    ascii_str32("encounter_name"),
    SIZE=84
    )

line = Struct("line",
    Bool16("flags",
        "addressee_look_at_speaker",
        "everyone_look_at_speaker",
        "everyone_look_at_addressee",
        "wait_until_told_to_advance",
        "wait_until_speaker_nearby",
        "wait_until_everyone_nearby",
        ),
    dyn_senum16("participant",
        DYN_NAME_PATH=".....participants.STEPTREE[DYN_I].encounter_name"),
    SEnum16("addressee",
        "none",
        "player",
        "participant",
        ),
    dyn_senum16("addressee_participant",
        DYN_NAME_PATH=".....participants.STEPTREE[DYN_I].encounter_name"),

    Pad(4),
    Float("line_delay_time"),

    Pad(12),
    dependency("variant_1", "snd!"),
    dependency("variant_2", "snd!"),
    dependency("variant_3", "snd!"),
    dependency("variant_4", "snd!"),
    dependency("variant_5", "snd!"),
    dependency("variant_6", "snd!"),
    SIZE=124
    )

ai_conversation = Struct("ai_conversation",
    ascii_str32("name"),
    Bool16("flags",
        "stop_if_death",
        "stop_if_damaged",
        "stop_if_visible_enemy",
        "stop_if_alerted_to_enemy",
        "player_must_be_visible",
        "stop_other_actions",
        "keep_trying_to_play",
        "player_must_be_looking",
        ),

    Pad(2),
    float_wu("trigger_distance"),
    float_wu("run_to_player_distance"),

    Pad(36),
    reflexive("participants", participant, 8,
        DYN_NAME_PATH='.encounter_name'),
    reflexive("lines", line, 32),
    SIZE=116
    )

structure_bsp = Struct("structure_bsp",
    UInt32("bsp_pointer", VISIBLE=False),
    UInt32("bsp_size", VISIBLE=False),
    UInt32("bsp_magic", VISIBLE=False),
    Pad(4),
    dependency("structure_bsp", "sbsp"),
    SIZE=32
    )

scnr_body = Struct("tagdata",
    dependency("DONT_USE", 'sbsp'),
    dependency("WONT_USE", 'sbsp'),
    dependency("CANT_USE", 'sky '),
    reflexive("skies", sky, 8, DYN_NAME_PATH='.sky.filepath'),
    SEnum16("type",
        "singleplayer",
        "multiplayer",
        "main_menu"
        ),
    Bool16("flags",
        "cortana_hack",
        "use_demo_ui"
        ),
    reflexive("child_scenarios", child_scenario, 16,
        DYN_NAME_PATH='.child_scenario.filepath'),
    float_rad("local_north"),  # radians

    Pad(156),
    reflexive("predicted_resources", predicted_resource, 1024, VISIBLE=False),
    reflexive("functions", function, 32,
        DYN_NAME_PATH='.name'),
    rawdata_ref("scenario_editor_data", max_size=65536),
    reflexive("comments", comment, 1024),

    Pad(224),
    reflexive("object_names", object_name, 512,
        DYN_NAME_PATH='.name', IGNORE_SAFE_MODE=True),
    reflexive("sceneries", scenery, 2000, IGNORE_SAFE_MODE=True),
    reflexive("sceneries_palette", scenery_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("bipeds", biped, 128, IGNORE_SAFE_MODE=True),
    reflexive("bipeds_palette", biped_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("vehicles", vehicle, 80, IGNORE_SAFE_MODE=True),
    reflexive("vehicles_palette", vehicle_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("equipments", equipment, 256, IGNORE_SAFE_MODE=True),
    reflexive("equipments_palette", equipment_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("weapons", weapon, 128, IGNORE_SAFE_MODE=True),
    reflexive("weapons_palette", weapon_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),

    reflexive("device_groups", device_group, 128,
        DYN_NAME_PATH='.name', IGNORE_SAFE_MODE=True),
    reflexive("machines", machine, 400, IGNORE_SAFE_MODE=True),
    reflexive("machines_palette", machine_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("controls", control, 100, IGNORE_SAFE_MODE=True),
    reflexive("controls_palette", control_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("light_fixtures", light_fixture, 500, IGNORE_SAFE_MODE=True),
    reflexive("light_fixtures_palette", light_fixture_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("sound_sceneries", sound_scenery, 256, IGNORE_SAFE_MODE=True),
    reflexive("sound_sceneries_palette", sound_scenery_swatch, 100,
        DYN_NAME_PATH='.name.filepath'),

    Pad(84),
    reflexive("player_starting_profiles", player_starting_profile, 256,
        DYN_NAME_PATH='.name'),
    reflexive("player_starting_locations", player_starting_location, 256),
    reflexive("trigger_volumes", trigger_volume, 256,
        DYN_NAME_PATH='.name'),
    reflexive("recorded_animations", recorded_animation, 1024,
        DYN_NAME_PATH='.name', IGNORE_SAFE_MODE=True),
    reflexive("netgame_flags", netgame_flag, 200,
        DYN_NAME_PATH='.type.enum_name'),
    reflexive("netgame_equipments", netgame_equipment, 200,
        DYN_NAME_PATH='.item_collection.filepath'),
    reflexive("starting_equipments", starting_equipment, 200),
    reflexive("bsp_switch_trigger_volumes", bsp_switch_trigger_volume, 256),
    reflexive("decals", decal, 65535),
    reflexive("decals_palette", decal_swatch, 128,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("detail_object_collection_palette",
        detail_object_collection_swatch, 32, DYN_NAME_PATH='.name.filepath'),

    Pad(84),
    reflexive("actors_palette", actor_swatch, 64,
        DYN_NAME_PATH='.name.filepath'),
    reflexive("encounters", encounter, 128, DYN_NAME_PATH='.name'),
    reflexive("command_lists", command_list, 256, DYN_NAME_PATH='.name'),
    reflexive("ai_animation_references", ai_anim_reference, 128,
        DYN_NAME_PATH='.animation_name'),
    reflexive("ai_script_references", ai_script_reference, 128,
        DYN_NAME_PATH='.script_name'),
    reflexive("ai_recording_references", ai_recording_reference, 128,
        DYN_NAME_PATH='.recording_name'),
    reflexive("ai_conversations", ai_conversation, 128,
        DYN_NAME_PATH='.name'),
    rawdata_ref("script_syntax_data", max_size=380076, IGNORE_SAFE_MODE=True),
    rawdata_ref("script_string_data", max_size=262144, IGNORE_SAFE_MODE=True),
    reflexive("scripts", halo_script, 512, DYN_NAME_PATH='.name'),
    reflexive("globals", halo_global, 128, DYN_NAME_PATH='.name'),
    reflexive("references", reference, 256,
              DYN_NAME_PATH='.reference.filepath'),
    reflexive("source_files", source_file, 8, DYN_NAME_PATH='.source_name'),

    Pad(24),
    reflexive("cutscene_flags", cutscene_flag, 512, DYN_NAME_PATH='.name'),
    reflexive("cutscene_camera_points", cutscene_camera_point, 512,
        DYN_NAME_PATH='.name'),
    reflexive("cutscene_titles", cutscene_title, 64, DYN_NAME_PATH='.name'),
    Pad(12), # OS bsp_modifiers reflexive

    Pad(96),
    dependency("custom_object_names", 'ustr'),
    dependency("ingame_help_text", 'ustr'),
    dependency("hud_messages", 'hmt '),
    reflexive("structure_bsps", structure_bsp, 16,
        DYN_NAME_PATH='.structure_bsp.filepath'),
    SIZE=1456,
    )

def get():
    return scnr_def

scnr_def = TagDef("scnr",
    blam_header('scnr', 2),
    scnr_body,

    ext=".scenario", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

tag_reference = Struct("tag_reference",
    dependency_os("tag"),
    SIZE=16
    )

tagc_body = Struct("tagdata",
    reflexive("tag_references", tag_reference, 200),
    SIZE=12,
    )


def get():
    return tagc_def

tagc_def = TagDef("tagc",
    blam_header('tagc'),
    tagc_body,

    ext=".tag_collection", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
from traceback import format_exc

from reclaimer.sounds import constants, util, blam_sound_samples
from supyr_struct.defs.audio.wav import wav_def
from supyr_struct.util import is_path_empty


class BlamSoundPermutation:
    name = ""

    # permutation properties
    _source_sample_data = b''
    _source_compression = constants.COMPRESSION_PCM_16_LE
    _source_sample_rate = constants.SAMPLE_RATE_22K
    _source_encoding = constants.ENCODING_MONO

    # processed properties
    _processed_samples = ()

    def __init__(self, sample_data=b'',
                 compression=constants.COMPRESSION_PCM_16_LE,
                 sample_rate=constants.SAMPLE_RATE_22K,
                 encoding=constants.ENCODING_MONO, **kwargs):
        self.load_source_samples(
            sample_data, compression, sample_rate, encoding)

    @property
    def source_sample_data(self):
        return self._source_sample_data
    @property
    def source_compression(self):
        return self._source_compression
    @property
    def source_sample_rate(self):
        return self._source_sample_rate
    @property
    def source_encoding(self):
        return self._source_encoding

    @property
    def processed_samples(self):
        return self._processed_samples
    @property
    def compression(self):
        try:
            return self.processed_samples[0].compression
        except Exception:
            return self._source_compression
    @property
    def sample_rate(self):
        try:
            return self.processed_samples[0].sample_rate
        except Exception:
            return self._source_sample_rate
    @property
    def encoding(self):
        try:
            return self.processed_samples[0].encoding
        except Exception:
            return self._source_encoding

    def load_source_samples(self, sample_data, compression,
                            sample_rate, encoding):
        self._source_sample_data = sample_data
        self._source_compression = compression
        self._source_sample_rate = sample_rate
        self._source_encoding = encoding
        self._processed_samples = []

    def partition_samples(self, target_compression=None,
                          target_sample_rate=None, target_encoding=None,
                          chunk_size=None, **compressor_kwargs):
        if target_compression is None:
            target_compression = self.source_compression

        if target_sample_rate is None:
            target_sample_rate = self.source_sample_rate

        if target_encoding is None:
            target_encoding = self.source_encoding

        if self.source_compression == constants.COMPRESSION_OGG:
            raise NotImplementedError(
                "Cannot partition Ogg Vorbis samples.")
        elif (target_compression not in constants.PCM_FORMATS and
              target_compression != constants.COMPRESSION_IMA_ADPCM and
              target_compression != constants.COMPRESSION_XBOX_ADPCM and
              target_compression != constants.COMPRESSION_OGG):
            raise ValueError('Unknown compression type "%s"' % target_compression)
        elif target_encoding not in (constants.ENCODING_MONO,
                                     constants.ENCODING_STEREO):
            raise ValueError("Compression encoding must be mono or stereo.")
        elif target_sample_rate <= 0:
            raise ValueError("Sample rate must be greater than zero.")

        source_compression = self.source_compression
        source_sample_rate = self.source_sample_rate
        source_encoding = self.source_encoding
        source_sample_data = self.source_sample_data

        target_chunk_size = util.get_sample_chunk_size(
            target_compression, target_encoding, chunk_size)
        if (source_compression == target_compression and
            source_sample_rate == target_sample_rate and
            source_encoding == target_encoding and
            (source_compression in constants.PCM_FORMATS or
             source_compression == constants.COMPRESSION_IMA_ADPCM or
             source_compression == constants.COMPRESSION_XBOX_ADPCM)):
            # compressing to same settings and can split at target_chunk_size
            # because format has fixed compression ratio. recompression not
            # necessary. Just split source into pieces at target_chunk_size.

            source_chunk_size = target_chunk_size
        else:
            # decompress samples so we can partition to a
            # different compression/encoding/sample rate
            decompressor = blam_sound_samples.BlamSoundSamples(
                source_sample_data, 0, source_compression,
                source_sample_rate, source_encoding
                )
            source_compression = constants.DEFAULT_UNCOMPRESSED_FORMAT
            source_sample_rate = target_sample_rate
            source_encoding = target_encoding
            source_sample_data = decompressor.get_decompressed(
                source_compression, source_sample_rate, source_encoding)

            source_bytes_per_sample = util.get_block_size(
                source_compression, source_encoding)

            target_blocksize = util.get_block_size(
                target_compression, target_encoding)
            target_samples_per_block = util.get_samples_per_block(
                target_compression)

            target_blocks_per_chunk = target_chunk_size // target_blocksize
            target_samples_per_chunk = (target_samples_per_block *
                                        target_blocks_per_chunk)
            source_chunk_size = (source_bytes_per_sample *
                                 target_samples_per_chunk)

        self._processed_samples = []
        for i in range(0, len(source_sample_data), source_chunk_size):
            chunk = source_sample_data[i: i + source_chunk_size]
            sample_count = util.get_sample_count(
                chunk, source_compression, source_encoding)

            self.processed_samples.append(
                blam_sound_samples.BlamSoundSamples(
                    chunk, sample_count, source_compression,
                    source_sample_rate, source_encoding)
                )

    def generate_mouth_data(self):
        for samples in self.processed_samples:
            samples.generate_mouth_data()

    def compress_samples(self, compression, sample_rate=None, encoding=None,
                         chunk_size=None, **compressor_kwargs):
        self.partition_samples(
            compression, sample_rate, encoding,
            chunk_size, **compressor_kwargs)
        for samples in self.processed_samples:
            samples.compress(compression, sample_rate, encoding,
                             **compressor_kwargs)

    def get_concatenated_sample_data(self, target_compression=None,
                                     target_sample_rate=None,
                                     target_encoding=None):
        if target_compression is None:
            target_compression = self.source_compression

        if target_sample_rate is None:
            target_sample_rate = self.source_sample_rate

        if target_encoding is None:
            target_encoding = self.source_encoding

        assert target_encoding in constants.channel_counts

        if (target_compression != self.compression or
            target_sample_rate != self.sample_rate or
            target_encoding != self.encoding):
            # decompress processed samples to the target compression
            sample_data = b''.join(
                p.get_decompressed(
                    target_compression, target_sample_rate, target_encoding)
                for p in self.processed_samples)
        else:
            # join samples without decompressing
            compression = self.compression
            # make sure we're able to combine samples without decompressing
            for piece in self.processed_samples:
                if piece.compression != compression:
                    raise ValueError(
                        "Cannot combine differently compressed samples without decompressing.")
                elif piece.compression == constants.COMPRESSION_OGG:
                    raise ValueError(
                        "Cannot combine ogg samples without decompressing.")

            sample_data = b''.join(p.sample_data for p in self.processed_samples)

        return sample_data

    def get_concatenated_mouth_data(self):
        return b''.join(p.mouth_data for p in self.processed_samples)

    def regenerate_source(self):
        '''
        Regenerates an uncompressed, concatenated audio stream
        from the compressed samples. Use when loading a sound tag
        for re-compression, re-sampling, or re-encoding.
        '''
        # always regenerate to constants.DEFAULT_UNCOMPRESSED_FORMAT
        # because, technically speaking, that is highest sample depth
        # we can ever possibly see in Halo CE.
        self._source_sample_data = self.get_concatenated_sample_data(
            constants.DEFAULT_UNCOMPRESSED_FORMAT,
            self.sample_rate, self.encoding)
        self._source_compression = constants.DEFAULT_UNCOMPRESSED_FORMAT
        self._source_sample_rate = self.sample_rate
        self._source_encoding = self.encoding

    @staticmethod
    def create_from_file(filepath):
        try:
            new_perm = BlamSoundPermutation()
            new_perm.import_from_file(filepath)
        except Exception:
            print(format_exc())
            new_perm = None

        return new_perm

    def export_to_file(self, filepath_base, overwrite=False,
                       export_source=True, decompress=True):
        perm_chunks = []
        encoding = self.encoding
        sample_rate = self.sample_rate
        if export_source and self.source_sample_data:
            # export the source data
            perm_chunks.append(
                (self.compression, self.source_encoding, self.source_sample_data)
                )
            sample_rate = self.source_sample_rate
        elif self.processed_samples:
            # concatenate processed samples if source samples don't exist.
            # also, if compression is ogg, we have to decompress
            compression = self.compression
            if decompress or compression == constants.COMPRESSION_OGG:
                compression = constants.COMPRESSION_PCM_16_LE

            try:
                sample_data = self.get_concatenated_sample_data(
                    compression, sample_rate, encoding)
                if sample_data:
                    perm_chunks.append((compression, self.encoding, sample_data))
            except Exception:
                print("Could not decompress permutation pieces. Concatenating.")
                perm_chunks.extend(
                    (piece.compression, piece.encoding, piece.sample_data)
                    for piece in self.processed_samples
                    )

        i = -1
        wav_file = wav_def.build()
        for compression, encoding, sample_data in perm_chunks:
            i += 1
            filepath = Path(util.BAD_PATH_CHAR_REMOVAL.sub("_", str(filepath_base)))

            if is_path_empty(filepath):
                filepath = Path("unnamed")

            if len(perm_chunks) > 1:
                filepath = filepath.parent.joinpath(filepath.stem + "__%s" % i)

            # figure out if the sample data is already encapsulated in a
            # container, or if it'll need to be encapsulated in a wav file.
            is_container_format = True
            if compression == constants.COMPRESSION_OGG:
                ext = ".ogg"
            elif compression == constants.COMPRESSION_WMA:
                ext = ".wma"
            elif compression == constants.COMPRESSION_UNKNOWN:
                ext = ".bin"
            else:
                is_container_format = False
                ext = ".wav"

            if filepath.suffix.lower() != ext:
                filepath = filepath.with_suffix(ext)

            if not sample_data or (not overwrite and filepath.is_file()):
                continue

            if is_container_format:
                try:
                    filepath.parent.mkdir(exist_ok=True, parents=True)
                    with filepath.open("wb") as f:
                        f.write(sample_data)
                except Exception:
                    print(format_exc())

                continue

            wav_file.filepath = filepath

            wav_fmt = wav_file.data.wav_format
            wav_chunks = wav_file.data.wav_chunks
            wav_chunks.append(case="data")
            data_chunk = wav_chunks[-1]

            wav_fmt.fmt.data = constants.WAV_FORMAT_PCM
            wav_fmt.channels = constants.channel_counts.get(encoding, 1)
            wav_fmt.sample_rate = sample_rate

            samples_len = len(sample_data)
            if compression in constants.PCM_FORMATS:
                # one of the uncompressed pcm formats
                if util.is_big_endian_pcm(compression):
                    sample_data = util.convert_pcm_to_pcm(
                        sample_data, compression,
                        util.change_pcm_endianness(compression))

                sample_width = constants.sample_widths[compression]
                wav_fmt.bits_per_sample = sample_width * 8
                wav_fmt.block_align = sample_width * wav_fmt.channels
                wav_fmt.byte_rate = wav_fmt.sample_rate * wav_fmt.block_align
            elif compression == constants.COMPRESSION_IMA_ADPCM:
                # 16bit adpcm
                wav_fmt.fmt.data = constants.WAV_FORMAT_IMA_ADPCM
                wav_fmt.bits_per_sample = 16
                wav_fmt.block_align = constants.IMA_ADPCM_COMPRESSED_BLOCKSIZE * wav_fmt.channels
                wav_fmt.byte_rate = int(
                    (wav_fmt.sample_rate * wav_fmt.block_align /
                     (constants.IMA_ADPCM_DECOMPRESSED_BLOCKSIZE // 2))
                    )
            elif compression == constants.COMPRESSION_XBOX_ADPCM:
                # 16bit adpcm
                wav_fmt.fmt.data = constants.WAV_FORMAT_XBOX_ADPCM
                wav_fmt.bits_per_sample = 16
                wav_fmt.block_align = constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE * wav_fmt.channels
                wav_fmt.byte_rate = int(
                    (wav_fmt.sample_rate * wav_fmt.block_align /
                     (constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE // 2))
                    )
            else:
                print("Unknown compression method:", compression)
                continue

            data_chunk.data = sample_data
            wav_file.data.wav_header.filesize = 36 + samples_len

            wav_file.serialize(temp=False, backup=False)

    def import_from_file(self, filepath):
        filepath = Path(filepath)
        if not filepath.is_file():
            raise OSError('File "%s" does not exist. Cannot import.' % filepath)

        wav_file = wav_def.build(filepath=filepath)
        wav_header = wav_file.data.wav_header
        wav_format = wav_file.data.wav_format
        wav_chunks = wav_file.data.wav_chunks
        data_chunk = None
        for chunk in wav_chunks:
            if chunk.sig.enum_name == "data":
                data_chunk = chunk
                break

        if wav_header.riff_sig != wav_header.get_desc("DEFAULT", "riff_sig"):
            raise ValueError(
                "RIFF signature is invalid. Not a valid wav file.")
        elif wav_header.wave_sig != wav_header.get_desc("DEFAULT", "wave_sig"):
            raise ValueError(
                "WAVE signature is invalid. Not a valid wav file.")
        elif wav_format.sig != wav_format.get_desc("DEFAULT", "sig"):
            raise ValueError(
                "Format signature is invalid. Not a valid wav file.")
        elif data_chunk is None:
            raise ValueError("Data chunk not present. Not a valid wav file.")
        elif wav_format.fmt.data not in constants.ALLOWED_WAV_FORMATS:
            raise ValueError(
                'Invalid compression format "%s".' % wav_format.fmt.data)
        elif wav_format.channels not in (1, 2):
            raise ValueError(
                "Invalid number of channels. Must be 1 or 2, not %s." %
                wav_format.channels)
        elif wav_format.sample_rate == 0:
            raise ValueError(
                "Sample rate cannot be zero. Not a valid wav file")
        elif (wav_format.fmt.data == constants.WAV_FORMAT_PCM_FLOAT and
              wav_format.bits_per_sample != 32):
            raise ValueError(
                "Pcm float sample width must be 32, not %s." %
                wav_format.bits_per_sample)
        elif (wav_format.fmt.data == constants.WAV_FORMAT_PCM and
              wav_format.bits_per_sample not in (8, 16, 24, 32)):
            raise ValueError(
                "Pcm sample width must be 8, 16, 24, or 32, not %s." %
                wav_format.bits_per_sample)

        if data_chunk.data_size != len(data_chunk.data):
            print("Audio sample data length does not match available data "
                  "length. Sample data may be truncated.")

        if wav_format.block_align and (len(data_chunk.data) %
                                       wav_format.block_align):
            print("Audio sample data length not a multiple of block_align. "
                  "Sample data may be truncated.")

        if wav_format.channels == 2:
            encoding = constants.ENCODING_STEREO
        else:
            encoding = constants.ENCODING_MONO

        sample_data = data_chunk.data
        if wav_format.fmt.data == constants.WAV_FORMAT_PCM_FLOAT:
            sample_data = util.convert_pcm_float32_to_pcm_32(sample_data)
            compression = constants.COMPRESSION_PCM_32_LE
        else:
            sample_width = None
            if wav_format.fmt.data == constants.WAV_FORMAT_PCM:
                sample_width = wav_format.bits_per_sample // 8

            compression = constants.wav_format_mapping.get(
                (wav_format.fmt.data, sample_width))

        self.load_source_samples(
            sample_data, compression, wav_format.sample_rate, encoding)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

string_reference = Struct("string_reference",
    dependency_os("string_id", "sidy"),
    SInt32("english_offset"),
    SInt32("unused_offset_1"),
    SInt32("unused_offset_2"),
    SInt32("unused_offset_3"),
    SInt32("unused_offset_4"),
    SInt32("unused_offset_5"),
    SInt32("unused_offset_6"),
    SInt32("unused_offset_7"),
    SInt32("unused_offset_8"),
    SIZE=56
    )

unic_body = Struct("tagdata",
    reflexive("string_references", string_reference, 9216,
        DYN_NAME_PATH='.string_id.filepath'),
    Pad(12),
    rawtext_ref("string_data_utf8", StrUtf8, max_size=18874368),

    SIZE=80
    )

def get():
    return unic_def

unic_def = TagDef("unic",
    blam_header_os('unic', 0),
    unic_body,

    ext=".multilingual_unicode_string_list", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import sqrt
from struct import unpack, pack_into
from types import MethodType

from reclaimer.hek.defs.objs.tag import HekTag
from reclaimer.util.compression import compress_normal32, decompress_normal32
from reclaimer.util.matrices import quaternion_to_matrix, Matrix

# TODO: Make calc_internal_data recalculate the lod nodes, and remove that
# same function from model.model_compilation.compile_gbxmodel and replace
# it with a call to calc_internal_data. lod nodes are recalculated when
# tags are compiled into maps, but the functionality should still be here.
class ModeTag(HekTag):

    def calc_internal_data(self):
        '''
        For each node, this method recalculates the rotation matrix
        from the quaternion, and the translation to the root bone.
        '''
        HekTag.calc_internal_data(self)

        nodes = self.data.tagdata.nodes.STEPTREE
        for node in nodes:
            rotation = quaternion_to_matrix(*node.rotation)
            trans = Matrix([node.translation])*-1
            parent = None

            # add the parents translation to this ones
            if node.parent_node > -1:
                trans += Matrix([nodes[node.parent_node].translation_to_root])

            # rotate the trans_to_root by this node's rotation
            trans *= rotation

            # combine this nodes rotation with its parents rotation
            if parent is not None:
                this_trans = node.translation
                node.distance_from_parent = sqrt(
                    this_trans.x**2 + this_trans.y**2 + this_trans.z**2)

                parent_rot = Matrix([parent.rot_jj_kk,
                                     parent.rot_kk_ii,
                                     parent.rot_ii_jj])
                rotation *= parent_rot

            # apply the changes to the node
            node.translation_to_root[:] = trans[0][:]
            node.rot_jj_kk[:] = rotation[0]
            node.rot_kk_ii[:] = rotation[1]
            node.rot_ii_jj[:] = rotation[2]

    def compress_part_verts(self, geometry_index, part_index):
        part = self.data.tagdata.geometries.STEPTREE\
               [geometry_index].parts.STEPTREE[part_index]
        uncomp_verts_reflexive = part.uncompressed_vertices
        comp_verts_reflexive = part.compressed_vertices

        comp_norm = compress_normal32
        unpack_vert = MethodType(unpack, ">11f2hf")
        pack_vert_into = MethodType(pack_into, ">12s3I2h2bh")

        comp_verts = bytearray(b'\x00' * 32 * uncomp_verts_reflexive.size)
        uncomp_verts = uncomp_verts_reflexive.STEPTREE

        in_off = out_off = 0
        # compress each of the verts and write them to the buffer
        for i in range(uncomp_verts_reflexive.size):
            ni, nj, nk, bi, bj, bk, ti, tj, tk,\
                u, v, ni_0, ni_1, nw = unpack_vert(
                    uncomp_verts[in_off + 12: in_off + 64])

            # write the compressed data
            pack_vert_into(
                comp_verts, out_off,
                uncomp_verts[in_off: in_off + 12],
                comp_norm(ni, nj, nk),
                comp_norm(bi, bj, bk),
                comp_norm(ti, tj, tk),
                int(max(0, min(1, u))*32767.5),
                int(max(0, min(1, v))*32767.5),
                ni_0*3, ni_1*3, int(max(0, min(1, nw))*32767.5))
            in_off  += 68
            out_off += 32

        comp_verts_reflexive.STEPTREE = comp_verts

    def decompress_part_verts(self, geometry_index, part_index):
        part = self.data.tagdata.geometries.STEPTREE\
               [geometry_index].parts.STEPTREE[part_index]
        uncomp_verts_reflexive = part.uncompressed_vertices
        comp_verts_reflexive = part.compressed_vertices

        decomp_norm = decompress_normal32
        unpack_vert = MethodType(unpack, ">3I2h2bh")
        pack_vert_into = MethodType(pack_into, ">12s11f2h2f")

        uncomp_verts = bytearray(b'\x00' * 68 * comp_verts_reflexive.size)
        comp_verts = comp_verts_reflexive.STEPTREE

        in_off = out_off = 0
        # uncompress each of the verts and write them to the buffer
        for i in range(comp_verts_reflexive.size):
            n, b, t, u, v, ni_0, ni_1, nw = unpack_vert(
                comp_verts[in_off + 12: in_off + 32])
            # write the uncompressed data
            pack_vert_into(
                uncomp_verts, out_off,
                comp_verts[in_off: in_off + 12],
                *decomp_norm(n), *decomp_norm(b), *decomp_norm(t),
                u/32767.5, v/32767.5, ni_0 // 3, ni_1 // 3,
                nw/32767.5, 1.0 - nw/32767.5)
            in_off  += 32
            out_off += 68

        uncomp_verts_reflexive.STEPTREE = uncomp_verts



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

part_body = Struct("tagdata",
    Bool32("flags",
        "can_animate_backwards",
        "animation_stops_at_rest",
        "animation_starts_on_random_frame",
        "animate_once_per_frame",
        "dies_at_rest",
        "dies_on_contact_with_structure",
        "tint_from_diffuse_texture",
        "dies_on_contact_with_water",
        "dies_on_contact_with_air",
        "self_illuminated",
        "random_horizontal_mirroring",
        "random_vertical_mirroring",
        ),
    dependency("bitmap", "bitm"),
    dependency("physics", "pphy"),
    dependency("impact_effect", "foot",
        TOOLTIP="Marty traded his kids for this"),

    Pad(4),
    from_to_sec("lifespan"),
    float_sec("fade_in_time"),
    float_sec("fade_out_time"),

    dependency("collision_effect", valid_event_effects),
    dependency("death_effect", valid_event_effects),

    Struct("rendering",
        Float("minimum_size", SIDETIP="pixels"),
        FlSInt32("unknown0", VISIBLE=False),
        FlFloat("unknown1", VISIBLE=False),
        QStruct("radius_animation", INCLUDE=from_to),
        Pad(4),
        QStruct("animation_rate",
            Float("from", UNIT_SCALE=per_sec_unit_scale, GUI_NAME=''),
            Float("to",   UNIT_SCALE=per_sec_unit_scale),
            ORIENT='h', SIDETIP='frames/sec'
            ),
        Float("contact_deterioration"),
        Float("fade_start_size", SIDETIP="pixels"),
        Float("fade_end_size", SIDETIP="pixels"),

        Pad(4),
        SInt16("first_sequence_index"),
        SInt16("initial_sequence_count"),
        SInt16("looping_sequence_count"),
        SInt16("final_sequence_count"),

        Pad(12),
        SEnum16("orientation", *render_mode),

        Pad(38),
        FlUInt32("unknown2", VISIBLE=False),
        Bool16("shader_flags", *shader_flags),
        SEnum16("framebuffer_blend_function", *framebuffer_blend_functions),
        SEnum16("framebuffer_fade_mode", *render_fade_mode),
        Bool16("map_flags",
            "unfiltered"
            ),
        ),
    Pad(12),  # OS v4 shader extension padding
    Pad(16),

    #Secondary map
    Struct("secondary_map",
        dependency("bitmap", "bitm"),
        SEnum16("anchor", *render_anchor),
        Bool16("flags",
            "unfiltered"
            ),
        Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
        Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
        Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),
        QStruct("rotation_center", INCLUDE=xy_float),
        Pad(4),
        Float("zsprite_radius_scale"),
        ),

    SIZE=356,
    )


def get():
    return part_def

part_def = TagDef("part",
    blam_header("part", 2),
    part_body,

    ext=".particle", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.ustr import UstrTag
from supyr_struct.defs.tag_def import TagDef

string_data_struct = rawtext_ref("string", FlStrUTF16, max_size=32768)

ustr_body = Struct("tagdata",
    reflexive("strings", string_data_struct, 32767,
        DYN_NAME_PATH='.data'),
    SIZE=12,
    )


def get():
    return ustr_def

ustr_def = TagDef("ustr",
    blam_header('ustr'),
    ustr_body,

    ext=".unicode_string_list", endian=">", tag_cls=UstrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.scnr import *
from supyr_struct.util import desc_variant

player_starting_profile = Struct("player_starting_profile",
    ascii_str32("name"),
    float_zero_to_one("starting_health_modifier"),
    float_zero_to_one("starting_shield_modifier"),
    dependency("primary_weapon", "weap"),
    SInt16("primary_rounds_loaded"),
    SInt16("primary_rounds_total"),
    dependency("secondary_weapon", "weap"),
    SInt16("secondary_rounds_loaded"),
    SInt16("secondary_rounds_total"),
    SInt8("starting_frag_grenade_count", MIN=0),
    SInt8("starting_plasma_grenade_count", MIN=0),
    SInt8("starting_custom_2_grenade_count", MIN=0),
    SInt8("starting_custom_3_grenade_count", MIN=0),
    SIZE=104
    )

ai_anim_reference = Struct("ai_animation_reference",
    ascii_str32("animation_name"),
    dependency_os("animation_graph", ("antr", "magy")),
    SIZE=60
    )

reference = Struct("tag_reference",
    Pad(24),
    dependency_os("reference"),
    SIZE=40
    )

# copy the scnr_body and replace the descriptors for certain
# fields with ones that are tweaked for use with open sauce
scnr_body = desc_variant(
    scnr_body,
    ("DONT_USE", dependency_os("project_yellow_definitions", 'yelo')),
    ("player_starting_profiles",
     reflexive("player_starting_profiles",
        player_starting_profile, 256, DYN_NAME_PATH='.name')
     ),
    ("ai_animation_references",
     reflexive("ai_animation_references",
        ai_anim_reference, 128, DYN_NAME_PATH='.animation_name')
     ),
    ("script_syntax_data", rawdata_ref("script_syntax_data", max_size=570076, IGNORE_SAFE_MODE=True)),
    ("script_string_data", rawdata_ref("script_string_data", max_size=393216, IGNORE_SAFE_MODE=True)),
    ("references",
     reflexive("references",
        reference, 256, DYN_NAME_PATH='.reference.filepath')
     ),
    ("structure_bsps",
     reflexive("structure_bsps",
        structure_bsp, 32, DYN_NAME_PATH='.structure_bsp.filepath')
     )
    )

def get():
    return scnr_def

scnr_def = TagDef("scnr",
    blam_header('scnr', 2),
    scnr_body,

    ext=".scenario", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'read_jms', 'write_jms', )


import re
import traceback

from copy import deepcopy
from pathlib import Path

from reclaimer.model.constants import (
    JMS_VERSION_HALO_1,
    JMS_VERSION_HALO_2_8210,
    JMS_PERM_CANNOT_BE_RANDOMLY_CHOSEN_TOKEN,
    )
from reclaimer.util import (
    float_to_str, float_to_str_truncate, parse_jm_float, parse_jm_int,
    )
from .model import JmsModel
from .node import JmsNode
from .material import JmsMaterial
from .marker import JmsMarker
from .vertex import JmsVertex
from .triangle import JmsTriangle

# Test showing off the regex can be seen here:
# https://regex101.com/r/7PKpWi/1
JMS_V1_SPLIT_REGEX = re.compile(r'\n+\s*|\t+')
# https://regex101.com/r/ySgI1Z/1
JMS_V2_SPLIT_REGEX = re.compile(r'\n+(;.*)*\s*|\t+')

def read_jms(jms_string, stop_at="", perm_name=None):
    '''
    Converts a jms string into a JmsModel instance.
    '''
    jms_data = tuple(JMS_V1_SPLIT_REGEX.split(jms_string))

    version = jms_data[0].strip()
    if version == JMS_VERSION_HALO_1:
        # Halo 1
        return _read_jms_8200(jms_data, stop_at, perm_name)

    # after 8200, comments are allowed. the comment character
    # is a semicolon, and the line must start with it.
    # filter out any lines that start with a semicolon.

    jms_string = jms_string.lstrip("\ufeff")
    jms_data = tuple(JMS_V2_SPLIT_REGEX.split(jms_string))

    version = jms_data[0].strip()
    if version == JMS_VERSION_HALO_2_8210:
        # Halo 2
        return _read_jms_8210(jms_data, stop_at)
    else:
        print("Unknown JMS version '%s'" % version)
        return None


def _read_jms_8200(jms_data, stop_at="", perm_name=None):
    if perm_name is None:
        perm_name = "__unnamed"

    jms_model = JmsModel(perm_name)

    # Halo 1
    dat_i = 0

    try:
        jms_model.version = str(parse_jm_int(jms_data[dat_i]))
        dat_i += 1
    except Exception:
        print(traceback.format_exc())
        print("Could not read version number.")
        return jms_model

    try:
        jms_model.node_list_checksum = parse_jm_int(jms_data[dat_i])
        dat_i += 1
    except Exception:
        print(traceback.format_exc())
        print("Could not read node list checksum.")
        return jms_model

    stop = (stop_at == "nodes")
    if not stop:
        # read the nodes
        try:
            i = 0 # make sure i is defined in case of exception
            jms_model.nodes[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(jms_model.nodes)):
                jms_model.nodes[i] = JmsNode(
                    jms_data[dat_i], parse_jm_int(jms_data[dat_i+1]), parse_jm_int(jms_data[dat_i+2]),
                    parse_jm_float(jms_data[dat_i+3]), parse_jm_float(jms_data[dat_i+4]),
                    parse_jm_float(jms_data[dat_i+5]), parse_jm_float(jms_data[dat_i+6]),
                    parse_jm_float(jms_data[dat_i+7]), parse_jm_float(jms_data[dat_i+8]), parse_jm_float(jms_data[dat_i+9]),
                    )
                dat_i += 10
            JmsNode.setup_node_hierarchy(jms_model.nodes)
        except Exception:
            print(traceback.format_exc())
            print("Failed to read nodes.")
            del jms_model.nodes[i: ]
            stop = True

    stop |= (stop_at == "materials")
    if not stop:
        # read the materials
        try:
            i = 0 # make sure i is defined in case of exception
            jms_model.materials[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(jms_model.materials)):
                jms_model.materials[i] = JmsMaterial(jms_data[dat_i], jms_data[dat_i+1])
                dat_i += 2
        except Exception:
            print(traceback.format_exc())
            print("Failed to read materials.")
            del jms_model.materials[i: ]
            stop = True

    stop |= (stop_at == "markers")
    if not stop:
        # read the markers
        try:
            i = 0 # make sure i is defined in case of exception
            jms_model.markers[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(jms_model.markers)):
                jms_model.markers[i] = JmsMarker(
                    jms_data[dat_i], jms_model.name,
                    parse_jm_int(jms_data[dat_i+1]), parse_jm_int(jms_data[dat_i+2]),
                    parse_jm_float(jms_data[dat_i+3]), parse_jm_float(jms_data[dat_i+4]),
                    parse_jm_float(jms_data[dat_i+5]), parse_jm_float(jms_data[dat_i+6]),
                    parse_jm_float(jms_data[dat_i+7]), parse_jm_float(jms_data[dat_i+8]), parse_jm_float(jms_data[dat_i+9]),
                    parse_jm_float(jms_data[dat_i+10])
                    )
                dat_i += 11
        except Exception:
            print(traceback.format_exc())
            print("Failed to read markers.")
            del jms_model.markers[i: ]
            stop = True

    stop |= (stop_at == "regions")
    if not stop:
        # read the regions
        try:
            i = 0 # make sure i is defined in case of exception
            jms_model.regions[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(jms_model.regions)):
                jms_model.regions[i] = jms_data[dat_i]
                dat_i += 1
        except Exception:
            print(traceback.format_exc())
            print("Failed to read regions.")
            del jms_model.regions[i: ]
            stop = True

    stop |= (stop_at == "vertices")
    if not stop:
        # read the vertices
        try:
            i = 0 # make sure i is defined in case of exception
            jms_model.verts[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(jms_model.verts)):
                jms_model.verts[i] = JmsVertex(
                    parse_jm_int(jms_data[dat_i]),
                    parse_jm_float(jms_data[dat_i+1]), parse_jm_float(jms_data[dat_i+2]), parse_jm_float(jms_data[dat_i+3]),
                    # tool normalizes imported jms normals by clamping, so we have to clamp as well
                    min(1.0, max(-1.0, parse_jm_float(jms_data[dat_i+4]))),
                    min(1.0, max(-1.0, parse_jm_float(jms_data[dat_i+5]))),
                    min(1.0, max(-1.0, parse_jm_float(jms_data[dat_i+6]))),
                    parse_jm_int(jms_data[dat_i+7]), parse_jm_float(jms_data[dat_i+8]),
                    parse_jm_float(jms_data[dat_i+9]), parse_jm_float(jms_data[dat_i+10]), parse_jm_float(jms_data[dat_i+11])
                    )
                dat_i += 12
        except Exception:
            print(traceback.format_exc())
            print("Failed to read vertices.")
            del jms_model.verts[i: ]
            stop = True

    stop |= (stop_at == "triangles")
    if not stop:
        # read the triangles
        try:
            i = 0 # make sure i is defined in case of exception
            jms_model.tris[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(jms_model.tris)):
                jms_model.tris[i] = JmsTriangle(
                    parse_jm_int(jms_data[dat_i]), parse_jm_int(jms_data[dat_i+1]),
                    parse_jm_int(jms_data[dat_i+2]), parse_jm_int(jms_data[dat_i+3]), parse_jm_int(jms_data[dat_i+4]),
                    )
                dat_i += 5
        except Exception:
            print(traceback.format_exc())
            print("Failed to read triangles.")
            del jms_model.tris[i: ]
            stop = True

    # return (jms_models[name], )
    return jms_model


def _read_jms_8210(jms_data, stop_at=""):
    # Halo 2
    # NOTE: This function is incomplete. It will not fully work
    jms_models = {}

    version = str(parse_jm_int(jms_data[0]))
    dat_i = 1

    nodes = []
    markers = []
    regions = []
    all_verts = []
    all_tris = {}

    material_perm_names = []

    perm_regions_and_shaders = {}
    cached_region_shader_indices = []

    stop = (stop_at == "nodes")
    if not stop:
        # read the nodes
        try:
            i = 0 # make sure i is defined in case of exception
            nodes[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(nodes)):
                # TODO: make the marker positions relative. according to
                #       General_101, the h2 nodes use absolute transforms
                nodes[i] = JmsNode(
                    jms_data[dat_i], -1, -1,  # these will need to be calculated
                    parse_jm_float(jms_data[dat_i+1]), parse_jm_float(jms_data[dat_i+2]),
                    parse_jm_float(jms_data[dat_i+3]), parse_jm_float(jms_data[dat_i+4]),
                    parse_jm_float(jms_data[dat_i+5]), parse_jm_float(jms_data[dat_i+6]),
                    parse_jm_float(jms_data[dat_i+7]), parse_jm_int(jms_data[dat_i+8]),
                    )
                dat_i += 9
            JmsNode.setup_node_hierarchy(nodes, version)
        except Exception:
            print(traceback.format_exc())
            print("Failed to read nodes.")
            del nodes[i: ]
            stop = True

    stop |= (stop_at == "materials")
    if not stop:
        # read the materials
        try:
            i = 0 # make sure i is defined in case of exception
            material_perm_names[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            cached_region_shader_indices[:] = (None, ) * len(material_perm_names)
            dat_i += 1
            for i in range(len(material_perm_names)):
                shader_name = jms_data[dat_i]
                mat_parts = jms_data[dat_i + 1].split()

                perm_name = mat_parts[-2]
                region_name = mat_parts[-1]

                material_perm_names[i] = perm_name

                perm_info = perm_regions_and_shaders.setdefault(perm_name, {})
                perm_regions = perm_info.setdefault("regions", [])
                perm_shaders = perm_info.setdefault("shaders", [])
                if region_name not in perm_regions:
                    perm_regions.append(region_name)

                if shader_name not in perm_shaders:
                    perm_shaders.append(shader_name)

                cached_region_shader_indices[i] = (
                    perm_regions.index(region_name),
                    perm_shaders.index(shader_name)
                    )
                dat_i += 2

        except Exception:
            print(traceback.format_exc())
            print("Failed to read materials.")
            del material_perm_names[i: ]
            del cached_region_shader_indices[i: ]
            stop = True

    stop |= (stop_at == "markers")
    if not stop:
        # read the markers
        try:
            i = 0 # make sure i is defined in case of exception
            markers[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(markers)):
                markers[i] = JmsMarker(
                    jms_data[dat_i], "", -1, parse_jm_int(jms_data[dat_i+1]),
                    parse_jm_float(jms_data[dat_i+2]), parse_jm_float(jms_data[dat_i+3]),
                    parse_jm_float(jms_data[dat_i+4]), parse_jm_float(jms_data[dat_i+5]),
                    parse_jm_float(jms_data[dat_i+6]), parse_jm_float(jms_data[dat_i+7]), parse_jm_float(jms_data[dat_i+8]),
                    parse_jm_float(jms_data[dat_i+9])
                    )
                dat_i += 10
        except Exception:
            print(traceback.format_exc())
            print("Failed to read markers.")
            del markers[i: ]
            stop = True

    stop |= (stop_at == "instance_xrefs")
    if not stop:
        # read the instance xrefs
        try:
            i = 0 # make sure i is defined in case of exception
            instance_xrefs = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(instance_xrefs)):
                instance_xrefs[i] = (jms_data[dat_i], jms_data[dat_i+1])
                dat_i += 2
        except Exception:
            print(traceback.format_exc())
            print("Failed to read instance xrefs.")
            del instance_xrefs[i: ]
            stop = True

    stop |= (stop_at == "instance_markers")
    if not stop:
        # read the instance markers
        try:
            i = 0 # make sure i is defined in case of exception
            instance_markers = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(instance_markers)):
                # TODO: do something about the permutation name for each marker.
                # maybe make one copy of the markers for each permutation
                instance_markers[i] = JmsMarker(
                    jms_data[dat_i], "", -1,
                    parse_jm_float(jms_data[dat_i+3]), parse_jm_float(jms_data[dat_i+4]),
                    parse_jm_float(jms_data[dat_i+5]), parse_jm_float(jms_data[dat_i+6]),
                    parse_jm_float(jms_data[dat_i+7]), parse_jm_float(jms_data[dat_i+8]), parse_jm_float(jms_data[dat_i+9]),
                    )
                dat_i += 10
        except Exception:
            print(traceback.format_exc())
            print("Failed to read instance markers.")
            del instance_markers[i: ]
            stop = True

    stop |= (stop_at == "vertices")
    if not stop:
        # read the vertices
        try:
            i = 0 # make sure i is defined in case of exception
            all_verts[:] = (None, ) * parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(len(all_verts)):
                x, y, z = parse_jm_float(jms_data[dat_i]),   parse_jm_float(jms_data[dat_i+1]), parse_jm_float(jms_data[dat_i+2])
                #a, b, c = parse_jm_float(jms_data[dat_i+3]), parse_jm_float(jms_data[dat_i+4]), parse_jm_float(jms_data[dat_i+5])

                # tool normalizes imported jms normals in halo 1 by clamping, so I'm going to assume halo 2 does as well
                a = min(1.0, max(-1.0, parse_jm_float(jms_data[dat_i+3])))
                b = min(1.0, max(-1.0, parse_jm_float(jms_data[dat_i+4])))
                c = min(1.0, max(-1.0, parse_jm_float(jms_data[dat_i+5])))
                dat_i += 6

                node_influences = [(-1, 0)] * parse_jm_int(jms_data[dat_i])
                dat_i += 1
                for j in range(len(node_influences)):
                    node_influences[j] = (parse_jm_int(jms_data[dat_i]),
                                          parse_jm_float(jms_data[dat_i+1]))
                    dat_i += 2

                tex_coords = [(0, 0)] * parse_jm_int(jms_data[dat_i])
                dat_i += 1
                for j in range(len(tex_coords)):
                    tex_coords[j] = (parse_jm_float(jms_data[dat_i]),
                                     parse_jm_float(jms_data[dat_i+1]))
                    dat_i += 2

                if len(node_influences) > 1:
                    node_1, node_1_weight = node_influences[1]
                else:
                    node_1, node_1_weight = -1, 0.0

                all_verts[i] = JmsVertex(
                    node_influences[0][0],
                    x, y, z,
                    a, b, c, 0, 1, 0, 1, 0, 0,
                    node_1, node_1_weight,
                    tex_coords[0][0], tex_coords[0][1], 0,
                    list(node[0] for node in node_influences[2: ]),
                    list(node[1] for node in node_influences[2: ]),
                    list(tex_coords[2: ])
                    )
        except Exception:
            print(traceback.format_exc())
            print("Failed to read vertices.")
            del all_verts[i: ]
            stop = True

    stop |= (stop_at == "triangles")
    if not stop:
        # read the triangles
        try:
            i = 0 # make sure i is defined in case of exception
            for perm_name in material_perm_names:
                all_tris[perm_name] = []

            all_tri_count = parse_jm_int(jms_data[dat_i])
            dat_i += 1
            for i in range(all_tri_count):
                mat_index = parse_jm_int(jms_data[dat_i])
                region, shader = cached_region_shader_indices[mat_index]

                all_tris[material_perm_names[mat_index]].append(JmsTriangle(
                    region, shader,
                    parse_jm_int(jms_data[dat_i+1]),
                    parse_jm_int(jms_data[dat_i+2]),
                    parse_jm_int(jms_data[dat_i+3]),
                    ))

                dat_i += 4
        except Exception:
            print(traceback.format_exc())
            print("Failed to read triangles.")
            stop = True

    for perm_name in material_perm_names:
        jms_model = JmsModel()

        jms_model.name = perm_name
        jms_model.perm_name = perm_name
        jms_model.version = version
        jms_model.is_random_perm = False

        jms_model.node_list_checksum = 0
        jms_model.nodes = deepcopy(nodes)
        jms_model.markers = deepcopy(markers)
        jms_model.tris    = all_tris[perm_name]
        jms_model.regions = perm_regions_and_shaders[perm_name]["regions"]
        jms_model.materials = list(
            JmsMaterial(name) for name in
            perm_regions_and_shaders[perm_name]["shaders"]
            )

        for marker in jms_model.markers:
            # set the permutation name for each marker
            marker.permutation = perm_name

        if True or len(material_perm_names) > 1:
            # more than one perm. need to split into multiple and rebase
            vert_rebase_map = {}
            for tri in all_tris[perm_name]:
                tri.v0 = vert_rebase_map.setdefault(tri.v0, len(vert_rebase_map))
                tri.v1 = vert_rebase_map.setdefault(tri.v1, len(vert_rebase_map))
                tri.v2 = vert_rebase_map.setdefault(tri.v2, len(vert_rebase_map))

            jms_model.verts = list(
                all_verts[vert_rebase_map[i]] for i in range(len(vert_rebase_map))
                )
        else:
            # only one perm. no need to split and rebase
            jms_model.verts = all_verts

        jms_models[perm_name] = jms_model


    # TODO: Make this able to read spheres, boxes, capsules,
    #       convex shapes, ragdolls, hinges, car wheels,
    #       point-to-points, prismatics, and bounding spheres

    # TODO: make several jms_models from the parsed data

    jms_models = tuple(jms_models[name] for name in sorted(jms_models))
    # uncomment when pipeline is changed to handle read_jms returning a tuple of JmsModels
    # return jms_models
    return jms_models[0] if jms_models else None


def write_jms(filepath, jms_model, use_blitzkrieg_rounding=False):
    '''
    Writes a JmsModel to filepath.
    '''
    if use_blitzkrieg_rounding:
        to_str = lambda f: float_to_str_truncate(f, 6)
    else:
        to_str = float_to_str

    materials = jms_model.materials
    regions = jms_model.regions

    # If the path doesnt exist, create it
    filepath = Path(filepath)
    filepath.parent.mkdir(exist_ok=True, parents=True)

    if not regions:
        regions = ("__unnamed", )

    if not materials:
        materials = (JmsMaterial("__unnamed", "<none>"), )

    with filepath.open("w", encoding='latin1', newline="\r\n") as f:
        f.write("%s\n" % jms_model.version)
        f.write("%s\n" % int(jms_model.node_list_checksum))

        f.write("%s\n" % len(jms_model.nodes))
        for node in jms_model.nodes:
            f.write("%s\n%s\n%s\n%s\t%s\t%s\t%s\n%s\t%s\t%s\n" % (
                node.name[: 31], node.first_child, node.sibling_index,
                to_str(node.rot_i), to_str(node.rot_j),
                to_str(node.rot_k), to_str(node.rot_w),
                to_str(node.pos_x), to_str(node.pos_y), to_str(node.pos_z),
                )
            )

        f.write("%s\n" % len(materials))
        for mat in materials:
            f.write("%s\n%s\n" % (mat.name + mat.properties, mat.tiff_path))

        f.write("%s\n" % len(jms_model.markers))
        for marker in jms_model.markers:
            f.write("%s\n%s\n%s\n%s\t%s\t%s\t%s\n%s\t%s\t%s\n%s\n" % (
                marker.name[: 31], marker.region, marker.parent,
                to_str(marker.rot_i), to_str(marker.rot_j),
                to_str(marker.rot_k), to_str(marker.rot_w),
                to_str(marker.pos_x), to_str(marker.pos_y), to_str(marker.pos_z),
                to_str(marker.radius)
                )
            )

        f.write("%s\n" % len(regions))
        for region in regions:
            f.write("%s\n" % region[: 31])

        f.write("%s\n" % len(jms_model.verts))
        for vert in jms_model.verts:
            f.write("%s\n%s\t%s\t%s\n%s\t%s\t%s\n%s\n%s\n%s\n%s\n%s\n" % (
                vert.node_0,
                to_str(vert.pos_x),  to_str(vert.pos_y),  to_str(vert.pos_z),
                to_str(vert.norm_i), to_str(vert.norm_j), to_str(vert.norm_k),
                vert.node_1,
                to_str(vert.node_1_weight),
                to_str(vert.tex_u), to_str(vert.tex_v), to_str(vert.tex_w),
                )
            )

        f.write("%s\n" % len(jms_model.tris))
        for tri in jms_model.tris:
            f.write("%s\n%s\n%s\t%s\t%s\n" % (
                tri.region, tri.shader,
                tri.v0, tri.v1, tri.v2
                )
            )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.tag import HekTag

class WphiTag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)
        tagdata = self.data.tagdata
        tagdata.crosshair_types.data = 0

        for crosshair in tagdata.crosshairs.STEPTREE:
            tagdata.crosshair_types.data |= 1 << crosshair.crosshair_type.data



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
    Adapted from source files located here
    https://github.com/Halogen002/Flare-Qt

    My thanks go to Halogen002 for providing me with
    the information I needed to write this definition.
    I extended it to include xbox gametypes as well
'''

from reclaimer.common_descs import *
from reclaimer.misc.defs.objs.pc_gametype import PcGametypeTag

from supyr_struct.defs.tag_def import TagDef

def get(): return pc_gametype_def

def is_xbox_gametype(node=None, parent=None, **kwargs):
    if parent is None:
        return node.get_root().is_xbox
    return parent.get_root().is_xbox

##################################################
'''Shared enumerators, booleans, and bitstructs'''
##################################################

enum_off_on = UEnum32('',
    'off',
    'on',
    )
#why is is that two of the slayer specific booleans reverse the truthyness?
enum_on_off = UEnum32('',
    'on',
    'off',
    )

speed_with_ball = UEnum32('speed with ball',
    'slow',
    'normal',
    'fast',
    )

trait_with_ball = UEnum32('trait with ball',
    'none',
    'invisible',
    'extra damage',
    'damage resistance',
    )

trait_without_ball = UEnum32('trait without ball',
    'none',
    'invisible',
    'extra damage',
    'damage resistance',
    )

ball_type = UEnum32('ball type',
    'normal',
    'reverse tag',
    'juggernaut',
    )

race_order = UEnum8('order',
    'normal',
    'any order',
    'rally',
    )

race_points_used = UEnum32('points used',
    'minimum',
    'maximum',
    'sum'
    )

vehicle_spawn = BitStruct("vehicle spawn",
    UBitEnum('vehicle type',
        'default',
        'none',
        'warthogs',
        'ghosts',
        'scorpions',
        'rocket warthogs',
        'banshees',
        'gun turrets',
        'custom',
        SIZE=4,
        ),
    UBitInt('warthogs',  SIZE=3),
    UBitInt('ghosts',    SIZE=3),
    UBitInt('scorpions', SIZE=3),
    UBitInt('rocket warthogs', SIZE=3),
    UBitInt('banshees',  SIZE=3),
    UBitInt('gun turrets', SIZE=3),
    SIZE=4,
    )

player_settings = Bool32("player settings",
    'radar enabled',
    'friend on hud',
    'infinite grenades',
    'shields disabled',
    'invisible',
    'generic weapons',
    'enemies not on radar',
    )

game_type = UEnum32('game type',
    ('ctf', 1),
    ('slayer', 2),
    ('oddball', 3),
    ('king', 4),
    ('race', 5),
    DEFAULT=1
    )

objective_indicator = UEnum32('objective indicator',
    'motion tracker',
    'nav point',
    'none',
    )

weapon_type = UEnum32('weapon type',
    'default',
    'pistols',
    'rifles',
    'plasma rifles',
    'sniper',
    'no sniper',
    'rocket launchers',
    'shotguns',
    'short range',
    'human',
    'covenant',
    'classic',
    'heavy weapons',
    )

friendly_fire = UEnum32('friendly fire',
    'off',
    'on',
    'shields',
    'explosions',
    )

vehicle_type = UEnum32('vehicle type',
    'all',
    'none',
    'warthog',
    'ghost',
    'scorpion',
    )

##################################################
'''Structs for each of the different game types'''
##################################################

ctf_settings = Struct('ctf settings',
    #UEnum8('assault', INCLUDE=enum_off_on),
    #UInt8('unknown', VISIBLE=False),
    #UEnum8('flag must reset', INCLUDE=enum_off_on),
    #UEnum8('flag must be at home', INCLUDE=enum_off_on),

    # It looks better this way.
    # Make it a FlBool32 so it doesnt get screwed up for powerpc
    FlBool32('flags',
        ('assault', 1<<0),
        ('flag must reset', 1<<16),
        ('flag must be at home', 1<<24),
        ),
    UInt32('single_flag_time', SIDETIP="seconds", UNIT_SCALE=1/30),
    SIZE=28,
    )

slayer_settings = Struct('slayer settings',
    #UEnum8('death bonus',   INCLUDE=enum_on_off),
    #UEnum8('kill penalty',  INCLUDE=enum_on_off),
    #UEnum8('kill in order', INCLUDE=enum_off_on),
    # It looks better this way.
    # Make it a FlBool32 so it doesnt get screwed up for powerpc
    FlBool32('flags',
        ('no death bonus', 1<<0),
        ('no kill penalty', 1<<8),
        ('kill in order', 1<<16),
        ),
    SIZE=28,
    )

oddball_settings = Struct('oddball settings',
    UEnum8('random ball', INCLUDE=enum_off_on),
    Pad(3),
    speed_with_ball,
    trait_with_ball,
    trait_without_ball,
    ball_type,
    UInt32('ball_count', MIN=1, MAX=16),
    SIZE=28,
    )

king_settings = Struct('king settings',
    UEnum8('moving hill', INCLUDE=enum_off_on),
    SIZE=28,
    )

race_settings = Struct('race settings',
    race_order,
    Pad(3),
    race_points_used,
    SIZE=28,
    )

header_comment = \
'''  The length of an Xbox gametypes name is capped at 11
  characters, whereas a PC gametype name is capped at 23.

  Respawn times of 0 are "instant", which is still 3 seconds.
  Respawn times cap at 300 seconds

  Health can be anywhere between 50% and 400%

  For PC gametypes, the max spawn for each kind of vehicle is 7.

  The score limit minimum is 1, and is measured in different
  units depending on what gametype this is:
      CTF -------- flags
      Slayer ----- kills
      King ------- minutes
      Oddball ---- points/minutes
      Race ------- laps'''

xbox_gametype_header = Struct("gametype header",
    StrUtf16('name', SIZE=24),
    game_type,
    UEnum32('teamplay', INCLUDE=enum_off_on),
    player_settings,
    objective_indicator,
    UEnum32('odd man out', INCLUDE=enum_off_on),

    UInt32('respawn time growth', SIDETIP="seconds", UNIT_SCALE=1/30),
    UInt32('respawn time', MAX=300*30, SIDETIP="seconds", UNIT_SCALE=1/30),
    UInt32('respawn suicide penalty',  SIDETIP="seconds", UNIT_SCALE=1/30),
    UInt32('lives', SIDETIP='[0 == unlimited]'),
    Float('health',
        DEFAULT=0.5, MIN=0.5, MAX=4.0, UNIT_SCALE=100, SIDETIP="%"),
    UInt32('score limit', MIN=1, DEFAULT=1),
    weapon_type,
    vehicle_type,
    SIZE=76, COMMENT=header_comment
    )


pc_gametype_header = Struct("gametype header",
    StrUtf16('name', SIZE=48),
    game_type,
    UEnum32('teamplay', INCLUDE=enum_off_on),
    player_settings,
    objective_indicator,
    UEnum32('odd man out', INCLUDE=enum_off_on),

    UInt32('respawn time growth', SIDETIP="seconds", UNIT_SCALE=1/30),
    UInt32('respawn time', MAX=300*30, SIDETIP="seconds", UNIT_SCALE=1/30),
    UInt32('respawn suicide penalty',  SIDETIP="seconds", UNIT_SCALE=1/30),
    UInt32('lives', SIDETIP='[0 == unlimited]'),
    Float('health',
        DEFAULT=0.5, MIN=0.5, MAX=4.0, UNIT_SCALE=100, SIDETIP="%"),
    UInt32('score limit', MIN=1, DEFAULT=1),
    weapon_type,

    BitStruct('red vehicles',  INCLUDE=vehicle_spawn),
    BitStruct('blue vehicles', INCLUDE=vehicle_spawn),

    UInt32('vehicle respawn time', SIDETIP="seconds", UNIT_SCALE=1/30),
    friendly_fire,
    UInt32('respawn betrayal penalty', SIDETIP="seconds", UNIT_SCALE=1/30),
    UEnum32('auto team balance', INCLUDE=enum_off_on),
    UInt32('time limit', SIDETIP="seconds", UNIT_SCALE=1/30),
    SIZE=124, COMMENT=header_comment
    )

xbox_gametype_footer = Container('gametype footer',
    #20 byte hmac sha1 digest of the save file
    BytesRaw('hmac_sig', SIZE=20),
    Pad(388),
    VISIBLE=False
    )

pc_gametype_footer = Struct('gametype footer',
    UInt32('crc 32'),
    #its possible to make a gametype platform independent by keeping
    #a copy of the settings here as well in a bytearray buffer
    BytearrayRaw('hybrid settings', SIZE=28),
    Pad(32),
    UInt32('crc 32 ce'),
    Pad(7972),
    VISIBLE=False
    )


header_switch = Switch('gametype header',
    DEFAULT=pc_gametype_header,
    CASE=is_xbox_gametype,
    CASES={True: xbox_gametype_header},
    )

union_settings_comment = '''
After you change these settings you'll still need to go into the
header and choose this gametypes type(ctf, slayer, race, etc).
'''

settings = Union('gametype settings',
    CASE='.gametype_header.game_type.enum_name',
    CASES={
        'ctf':ctf_settings,
        'slayer':slayer_settings ,
        'oddball':oddball_settings,
        'king':king_settings,
        'race':race_settings,
        },
    COMMENT=union_settings_comment
    )

footer_switch = Switch('gametype footer',
    DEFAULT=pc_gametype_footer,
    CASE=is_xbox_gametype,
    CASES={True: xbox_gametype_footer},
    )


pc_gametype_def = TagDef('pc_gametype',
    header_switch,
    settings,
    footer_switch,

    ext='.lst', endian='<', tag_cls=PcGametypeTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.meta.halo_map import tag_index_pc_def


def h2_alpha_to_h1_tag_index(map_header, tag_index):
    new_index = tag_index_pc_def.build()
    old_index_array = tag_index.tag_index
    new_index_array = new_index.tag_index

    # copy information from the h2 index into the h1 index
    new_index.scenario_tag_id = tag_index.scenario_tag_id
    new_index.tag_index_offset = tag_index.tag_index_offset
    new_index.tag_count = tag_index.tag_count

    for i in range(len(old_index_array)):
        old_index_entry = old_index_array[i]
        new_index_array.append()
        new_index_entry = new_index_array[-1]

        new_index_entry.class_1 = old_index_entry.class_1
        new_index_entry.class_2 = old_index_entry.class_2
        new_index_entry.class_3 = old_index_entry.class_3

        new_index_entry.id  = old_index_entry.id
        new_index_entry.pad = old_index_entry.flags
        new_index_entry.path_offset = old_index_entry.path_offset
        new_index_entry.meta_offset = old_index_entry.meta_offset
        new_index_entry.path = old_index_entry.path

    return new_index


def h2_to_h1_tag_index(map_header, tag_index):
    new_index = tag_index_pc_def.build()
    old_index_array = tag_index.tag_index
    new_index_array = new_index.tag_index

    # copy information from the h2 index into the h1 index
    new_index.scenario_tag_id = tag_index.scenario_tag_id
    new_index.tag_index_offset = tag_index.tag_index_offset
    new_index.tag_count = tag_index.tag_count

    tag_types = {}
    for typ in tag_index.tag_types:
        tag_types[typ.class_1.data] = [typ.class_1, typ.class_2, typ.class_3]

    for i in range(len(old_index_array)):
        old_index_entry = old_index_array[i]
        new_index_array.append()
        new_index_entry = new_index_array[-1]
        if old_index_entry.tag_class.data not in tag_types:
            new_index_entry.path = "reserved"
            new_index_entry.class_1.data = new_index_entry.class_2.data =\
                                           new_index_entry.class_3.data =\
                                           0xFFFFFFFF
            new_index_entry.id = 0xFFFFFFFF
            continue

        types = tag_types[old_index_entry.tag_class.data]
        new_index_entry.class_1 = types[0]
        new_index_entry.class_2 = types[1]
        new_index_entry.class_3 = types[2]

        new_index_entry.path = map_header.strings.\
                               tag_name_table[i].tag_name

        new_index_entry.id = old_index_entry.id
        new_index_entry.meta_offset = old_index_entry.offset
        if new_index_entry.meta_offset == 0:
            # might flag sbsp and ltmp tags as indexed
            new_index_entry.indexed = 1

    return new_index


def h3_to_h1_tag_index(map_header, tag_index):
    new_index = tag_index_pc_def.build()
    old_index_array = tag_index.tag_index
    new_index_array = new_index.tag_index

    # copy information from the h2 index into the h1 index
    #new_index.scenario_tag_id = tag_index.scenario_tag_id
    new_index.tag_index_offset = tag_index.tag_index_offset
    new_index.tag_count = tag_index.tag_count

    tag_types = {}
    for i in range(len(tag_index.tag_types)):
        typ = tag_index.tag_types[i]
        tag_types[i] = [typ.class_1, typ.class_2, typ.class_3]

    for i in range(len(old_index_array)):
        old_index_entry = old_index_array[i]
        new_index_array.append()
        new_index_entry = new_index_array[-1]
        if old_index_entry.tag_type_index not in tag_types:
            new_index_entry.path = "reserved"
            new_index_entry.class_1.data = new_index_entry.class_2.data =\
                                           new_index_entry.class_3.data =\
                                           0xFFFFFFFF
            new_index_entry.id = 0xFFFFFFFF
            continue

        types = tag_types[old_index_entry.tag_type_index]
        new_index_entry.class_1 = types[0]
        new_index_entry.class_2 = types[1]
        new_index_entry.class_3 = types[2]

        new_index_entry.path = map_header.strings.\
                               tag_name_table[i].tag_name

        new_index_entry.id = (old_index_entry.table_index << 16) + i
        new_index_entry.meta_offset = old_index_entry.offset
        if new_index_entry.meta_offset == 0:
            # might flag sbsp and ltmp tags as indexed
            new_index_entry.indexed = 1

    return new_index



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.tag import HekTag

class Str_Tag(HekTag):

    def calc_internal_data(self):
        HekTag.calc_internal_data(self)
        strings = self.data.tagdata.strings.STEPTREE

        for i in range(len(strings)):
            # replace all instances of \r and \n with \r\n
            split_strings = []
            for s in strings[i].data.split("\r\n"):
                for sub_s in s.split('\r'):
                    split_strings.extend(sub_s.split('\n'))

            strings[i].data = "\r\n".join(split_strings)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

sgla_attrs = Struct("sgla_attrs",
    #Glass Shader Properties
    Bool16("glass_shader_flags",
        "alpha_tested",
        "decal",
        "two_sided",
        "bump_map_is_specular_mask",
        ),

    Pad(42),
    #Background Tint Properties
    Struct("background_tint_properties",
        QStruct("color", INCLUDE=rgb_float),
        Float("map_scale"),
        dependency("map", "bitm"),
        ),

    Pad(22),
    #Reflection Properties
    Struct("reflection_properties",
        SEnum16("type",
            "bumped_cubemap",
            "flat_cubemap",
            "dynamic_mirror",
            ),
        float_zero_to_one("perpendicular_brightness"),  # [0,1]
        QStruct("perpendicular_tint_color", INCLUDE=rgb_float),
        float_zero_to_one("parallel_brightness"),  # [0,1]
        QStruct("parallel_tint_color", INCLUDE=rgb_float),
        dependency("map", "bitm"),

        Float("bump_map_scale"),
        dependency("bump_map", "bitm"),
        ),

    Pad(132),
    #Diffuse Properties
    Struct("diffuse_properties",
        Float("map_scale"),
        dependency("map", "bitm"),
        Float("detail_map_scale"),
        dependency("detail_map", "bitm"),
        ),

    Pad(32),
    #Specular Properties
    Struct("specular_properties",
        Float("map_scale"),
        dependency("map", "bitm"),
        Float("detail_map_scale"),
        dependency("detail_map", "bitm"),
        ),
    SIZE=440
    )

sgla_body = Struct("tagdata",
    shdr_attrs,
    sgla_attrs,
    SIZE=480,
    )


def get():
    return sgla_def

sgla_def = TagDef("sgla",
    blam_header('sgla'),
    sgla_body,

    ext=".shader_transparent_glass", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.part import *

part_body = dict(part_body)
part_body[11] = reflexive(
    "particle_shader_extensions",
    Struct("particle_shader_extension", INCLUDE=os_shader_extension),
    1)


def get():
    return part_def

part_def = TagDef("part",
    blam_header("part", 2),
    part_body,

    ext=".particle", endian=">", tag_cls=HekTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h3.common_descs import *
from supyr_struct.defs.tag_def import BlockDef


def compute_partition_offset(parent, **kwargs):
    if parent.size == 0:
        parent.file_offset = 0
        return

    partitions = parent.parent
    partition_num = partitions.index(parent)

    sections = partitions.parent.sections
    file_offset = 4096*3
    for i in range(2): # tag data is the third section
        file_offset += sections[i].size

    for i in range(partition_num):
        file_offset += partitions[i].size

    parent.file_offset = file_offset


def compute_section_offset(parent, **kwargs):
    if parent.size == 0:
        parent.file_offset = 0
        return

    sections = parent.parent
    section_num = sections.index(parent)

    file_offset = 4096*3
    for i in range(section_num):
        file_offset += sections[i].size

    parent.file_offset = file_offset


def virtual_ptr_to_file_ptr(ptr, section=0, map_header=None, **kwargs):
    if not map_header:
        return ptr
    elif section in (0, 1, 3):
        for section in map_header.sections:
            if ptr in range(section.virtual_address,
                            section.virtual_address + section.size):
                return ptr - section.virtual_address + section.file_offset
    else:
        for partition in map_header.partitions:
            if ptr in range(partition.load_address,
                            partition.load_address + partition.size):
                return ptr - partition.load_address + partition.file_offset
    raise ValueError("Pointer '%s' is not in any partitions." % ptr)


def file_ptr_to_virtual_ptr(ptr, section=0, map_header=None, **kwargs):
    if not map_header:
        return ptr
    elif section in (0, 1, 3):
        for section in map_header.sections:
            if ptr in range(section.file_offset,
                            section.file_offset + section.size):
                return ptr - section.file_offset + section.virtual_address
    else:
        for partition in map_header.partitions:
            if ptr in range(partition.file_offset,
                            partition.file_offset + partition.size):
                return ptr - partition.file_offset + partition.load_address
    raise ValueError("Pointer '%s' is not in any partitions." % ptr)


def string_id_table_name_pointer(parent=None, new_value=None, **kwargs):
    start = parent.parent.parent.string_id_table_file_offset
    if new_value is None:
        if parent.offset == -1:
            return start
        return parent.offset + start
    parent.offset = new_value - start

def tag_name_table_name_pointer(parent=None, new_value=None, **kwargs):
    start = parent.parent.parent.tag_name_table_file_offset
    if new_value is None:
        if parent.offset == -1:
            return start
        return parent.offset + start
    parent.offset = new_value - start


def compute_string_id_index_offset(parent, **kw):
    parent.string_id_index_file_offset = virtual_ptr_to_file_ptr(
        parent.get_root().string_id_index_offset,
        0, map_header=parent.get_root())

def compute_string_id_table_offset(parent, **kw):
    parent.string_id_table_file_offset = virtual_ptr_to_file_ptr(
        parent.get_root().string_id_table_offset,
        0, map_header=parent.get_root())

def compute_tag_name_index_offset(parent, **kw):
    parent.tag_name_index_file_offset = virtual_ptr_to_file_ptr(
        parent.get_root().tag_name_index_offset,
        0, map_header=parent.get_root())

def compute_tag_name_table_offset(parent, **kw):
    parent.tag_name_table_file_offset = virtual_ptr_to_file_ptr(
        parent.get_root().tag_name_table_offset,
        0, map_header=parent.get_root())


def tag_types_array_pointer(parent=None, new_value=None, **kwargs):
    if new_value is None:
        return virtual_ptr_to_file_ptr(parent.tag_types_offset, 2, **kwargs)
    parent.tag_types_offset = file_ptr_to_virtual_ptr(new_value, **kwargs)

def tag_index_array_pointer(parent=None, new_value=None, **kwargs):
    if new_value is None:
        return virtual_ptr_to_file_ptr(parent.tag_index_offset, 2, **kwargs)
    parent.tag_index_offset = file_ptr_to_virtual_ptr(new_value, **kwargs)

def root_tags_array_pointer(parent=None, new_value=None, **kwargs):
    if new_value is None:
        return virtual_ptr_to_file_ptr(parent.root_tags_offset, 2, **kwargs)
    parent.root_tags_offset = file_ptr_to_virtual_ptr(new_value, **kwargs)


string_id_table_entry = Container("string id table entry",
    SInt32("offset"),
    STEPTREE=CStrLatin1("string", POINTER=string_id_table_name_pointer),
    )

tag_name_table_entry = Container("tag name table entry",
    SInt32("offset"),
    STEPTREE=CStrLatin1("tag name", POINTER=tag_name_table_name_pointer),
    )


string_id_table = Array("string id table",
    SIZE="..string_id_count", SUB_STRUCT=string_id_table_entry,
    POINTER=".string_id_index_file_offset",
    )

tag_name_table = Array("tag name table",
    SIZE="..tag_name_count", SUB_STRUCT=tag_name_table_entry,
    POINTER=".tag_name_index_file_offset"
    )

partition = Struct("partition",
    UInt32("load address"),
    UInt32("size"),
    STEPTREE=Computed("file offset",
        COMPUTE_READ=compute_partition_offset, WIDGET_WIDTH=10
        ),
    SIZE=8
    )

section = Struct("section",
    UInt32("virtual address"),
    UInt32("size"),
    Computed("file offset",
        COMPUTE_READ=compute_section_offset, WIDGET_WIDTH=10
        ),
    SIZE=8
    )

h3_map_header = Struct("map header",
    UEnum32('head', ('head', 'head'),
        EDITABLE=False, VISIBLE=False, DEFAULT='head'
        ),
    UEnum32("version",
        ("halo1xbox",   5),
        ("halo1pcdemo", 6),
        ("halo1pc", 7),
        ("halo2", 8),
        ("halo3beta", 9),
        ("halo3", 11),
        ("haloreach", 12),
        ("halo1ce", 609),
        ),
    UInt32("decomp len"),
    UInt32("unknown0", VISIBLE=False),
    UInt32("tag index header offset"),
    UInt32("tag buffer offset"),
    UInt32("virtual size"),
    Pad(256),

    ascii_str32("build date"),
    UEnum16("map type",
        "sp",
        "mp",
        "ui",
        "shared",
        "sharedsp",
        ),
    UInt16("unknown2", VISIBLE=False),
    UInt8("unknown3", VISIBLE=False),
    UInt8("unknown4", VISIBLE=False),
    UInt8("unknown5", VISIBLE=False),
    UInt8("unknown6", VISIBLE=False),
    Pad(20),
    UInt32("string id count"),
    UInt32("string id table size"),
    UInt32("string id index offset"),
    UInt32("string id table offset"),

    Bool8("required maps",
        "ui",
        "shared",
        "sp_shared"
        ),
    Pad(3),
    UInt32("build_date_low"),
    UInt32("build_date_high"),
    UInt32("unknown_ui_map_related_0"),
    UInt32("unknown_ui_map_related_1"),
    UInt32("unknown_shared_map_related_0"),
    UInt32("unknown_shared_map_related_1"),
    UInt32("unknown_sp_shared_map_related_0"),
    UInt32("unknown_sp_shared_map_related_1"),
    ascii_str32("map name"),
    Pad(4),
    StrLatin1("scenario tag path", SIZE=260),
    UInt32("tag name count"),
    UInt32("tag name table offset"),
    UInt32("tag name table size"),
    UInt32("tag name index offset"),

    UInt32("crc32"),
    BytesRaw("unknown9", SIZE=32, VISIBLE=False),
    UInt32("virtual address"),
    UInt32("xkd version"),
    Array("partitions", SUB_STRUCT=partition, SIZE=6),

    UInt32("unknown_count"),
    BytesRaw("unknown10", SIZE=12, VISIBLE=False),
    BytesRaw("unknown11", SIZE=316, VISIBLE=False),
    Array("offset masks", SUB_STRUCT=UInt32("mask"), SIZE=4,
        NAME_MAP={"debug": 0, "resource": 1, "tag": 2, "locale": 3}
        ),
    Array("sections", SUB_STRUCT=section, SIZE=4,
        NAME_MAP={"debug": 0, "resource": 1, "tag": 2, "locale": 3}
        ),

    Pad(11104),
    UEnum32('foot', ('foot', 'foot'), EDITABLE=False, DEFAULT='foot'),
    STEPTREE=Container("strings",
        Computed("string id index file offset",
            COMPUTE_READ=compute_string_id_index_offset, WIDGET_WIDTH=10
            ),
        Computed("string id table file offset",
            COMPUTE_READ=compute_string_id_table_offset, WIDGET_WIDTH=10
            ),
        Computed("tag name index file offset",
            COMPUTE_READ=compute_tag_name_index_offset, WIDGET_WIDTH=10
            ),
        Computed("tag name table file offset",
            COMPUTE_READ=compute_tag_name_table_offset, WIDGET_WIDTH=10
            ),
        string_id_table,
        tag_name_table,
        ),
    SIZE=4096*3, ENDIAN=">"
    )

h3_tag_type = Struct("tag type",
    UEnum32("class 1", GUI_NAME="primary tag class",   INCLUDE=valid_h3_tags),
    UEnum32("class 2", GUI_NAME="secondary tag class", INCLUDE=valid_h3_tags),
    UEnum32("class 3", GUI_NAME="tertiary tag class",  INCLUDE=valid_h3_tags),
    h3_string_id("string id"),  # specifies the string containing the file
    #                             extension for this class(i.e: bitm = bitmap)
    )

h3_tag_header = Struct("tag header",
    UInt16("tag type index"),
    UInt16("table index"),
    UInt32("offset"),
    SIZE=8
    )

h3_root_tag_ref = Struct("root tag ref",
    UEnum32("tag class", INCLUDE=valid_h3_tags),
    UInt32("id"),
    SIZE=8
    )

h3_tag_types_array = Array("tag types",
    SIZE=".tag_types_count", SUB_STRUCT=h3_tag_type,
    POINTER=tag_types_array_pointer
    )

h3_tag_index_array = TagIndex("tag index",
    SIZE=".tag_count", SUB_STRUCT=h3_tag_header,
    POINTER=tag_index_array_pointer
    )

h3_root_tags_array = TagIndex("root tags",
    SIZE=".root_tags_count", SUB_STRUCT=h3_root_tag_ref,
    POINTER=root_tags_array_pointer
    )

h3_tag_index = Container("tag index",
    UInt32("tag types count"),
    UInt32("tag types offset"),

    UInt32("tag count"),
    UInt32("tag index offset"),

    UInt32("root tags count"),
    UInt32("root tags offset"),
    UInt32("unknown count"),
    UInt32("unknown offset"),

    UInt32("crc32"),
    UInt32("tag sig", EDITABLE=False, DEFAULT='tags'),

    h3_tag_types_array,
    h3_tag_index_array,
    h3_root_tags_array,
    ENDIAN=">"
    )

h3_map_header_def = BlockDef(h3_map_header)
h3_tag_index_def = BlockDef(h3_tag_index)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
The formula for calculating the magic for ANY map is as follows:
    magic = idx_magic - idx_off

idx_off is the offset of the tag_index_header and idx_magic
varies depending on which engine version the map is for.
To get the idx_magic, do:
    map_magics[map_header.version.enum_name]

where map_magics is in reclaimer.constants

To convert a magic relative pointer to an absolute pointer,
simply do:    abs_pointer = magic_pointer - magic
'''

from reclaimer.common_descs import *

from supyr_struct.defs.block_def import BlockDef
from supyr_struct.util import desc_variant

def tag_path_pointer(parent=None, new_value=None, magic=0, **kwargs):
    '''
    Calculates the pointer value for a tag_path based on the file offset.
    '''
    if parent is None:
        raise KeyError()
    if new_value is None:
        return parent.path_offset - magic
    parent.path_offset = new_value + magic


def tag_index_array_pointer(parent=None, new_value=None, magic=0, **kwargs):
    '''
    Calculates the pointer value for a tag_index array based on file offset.
    '''
    if new_value is None:
        return parent.tag_index_offset - magic
    parent.tag_index_offset = new_value + magic


yelo_header = Struct("yelo header",
    UEnum32("yelo", ('yelo', 'yelo'), EDITABLE=False, DEFAULT='yelo'),
    UEnum16("version type",
        ("version", 1),
        ("version minimum build", 2)
        ),
    Bool16("flags",
        "uses memory upgrades",
        "uses mod data files",
        "is protected",
        "uses game state upgrades",
        "has compression params",
        ),
    QStruct("tag versioning",
        UInt8("project yellow"),
        UInt8("project yellow globals"),
        Pad(2),
        SIZE=4
        ),

    Float("memory upgrade multiplier"),

    Struct("cheape definitions",
        UInt32("size"),
        UInt32("decompressed size"),
        UInt32("offset"),
        Pad(4),

        ascii_str32("build string"),
        SIZE=48
        ),

    ascii_str32("mod name"),

    Struct("build info",
        Pad(2),
        UEnum16("stage",
            "ship",
            "alpha",
            "beta",
            "delta",
            "epsilon",
            "release",
            ),
        UInt32("revision"),
        Timestamp64("timestamp"),

        ascii_str32("build string"),

        QStruct("cheape",
            UInt8("maj"),
            UInt8("min"),
            UInt16("build"),
            ),
        BytesRaw("uuid buffer", SIZE=16),

        QStruct("minimum os build",
            UInt8("maj"),
            UInt8("min"),
            UInt16("build"),
            ),
        Pad(4*3),
        SIZE=84
        ),

    QStruct("resources",
        UInt32("compression params header offset"),
        UInt32("tag symbol storage header offset"),
        UInt32("string id storage header offset"),
        UInt32("tag string to id storage header offset"),
        SIZE=16
        ),

    SIZE=196
    )

vap_block = Struct("vap_block",
    # File offset of the compressed block
    UInt32("file_offset"),

    # File size of the compressed block
    UInt32("file_size"),

    # Decompressed size of the compressed block
    UInt32("decompressed_size"),

    # Reserved for potential future usage
    UInt32("reserved"),
    )

# validated archive pattern(uh huh, sure)
vap_header = Struct("vap_header",
    # File size of the map when decompressed, include header size
    UInt32("decompressed_size"),
    UEnum16("compression_type",
        "uncompressed",
        "lzma",
    ),
    UEnum16("vap_version",
        "chimera_1",
        "next",
    ),

    # Number of blocks. If 0, data is assumed to be one contiguous stream
    UInt32("block_count"),
    # File size of the map when compressed, include header size
    UInt32("compressed_size"),

    UEnum32("feature_level",
        "chimera_1",
        "next",
    ),

    # Player limit (campaign). Set to 0 for multiplayer and user interface maps
    UInt16("max_players"),
    # Reserved for future use. Leave it at 0
    Pad(10),

    # Build date of the map in ISO 8601 (yyyy-mm-ddThh:mm:ss.fffffffff) format
    ascii_str32("build_date"),

    # Human-readable of the map
    ascii_str32("name"),

    # Human-readable description of the map
    StrLatin1("description", SIZE=128),
    Pad(256),
    STEPTREE=Array("blocks",
        SIZE=".block_count", SUB_STRUCT=vap_block
        )
    )


# Halo Demo maps have a different header
# structure with garbage filling the padding
map_header_demo = Struct("map header",
    Pad(2),
    UEnum16("map type",
        "sp",
        "mp",
        "ui",
        ),
    Pad(700),
    UEnum32('head', ('head', 'Ehed'), EDITABLE=False, DEFAULT='Ehed'),
    UInt32("tag data size"),
    ascii_str32("build date", EDITABLE=False),
    Pad(672),
    UEnum32("version",
        ("halo1xbox",   5),
        ("halo1pcdemo", 6),
        ("halo1pc", 7),
        ("halo2", 8),
        ("halo3beta", 9),
        ("halo3", 11),
        ("halo1ce", 609),
        ("halo1vap", 134),
        ),
    ascii_str32("map name"),
    UInt32("unknown"),
    UInt32("crc32"),
    Pad(52),
    UInt32("decomp len"),
    UInt32("tag index header offset"),
    UEnum32('foot', ('foot', 'Gfot'), EDITABLE=False, DEFAULT='Gfot'),
    Pad(524),
    SIZE=2048
    )

map_header = Struct("map header",
    UEnum32('head', ('head', 'head'), DEFAULT='head'),
    UEnum32("version",
        ("halo1xbox",   5),
        ("halo1pcdemo", 6),
        ("halo1pc", 7),
        ("halo2", 8),
        ("halo3beta", 9),
        ("halo3", 11),
        ("halo1ce", 609),
        ("halo1vap", 134),
        ),
    UInt32("decomp len"),
    UInt32("unknown"),
    UInt32("tag index header offset"),
    UInt32("tag data size"),
    Pad(8),
    ascii_str32("map name"),
    ascii_str32("build date", EDITABLE=False),
    UEnum16("map type",
        "sp",
        "mp",
        "ui",
        ),
    Pad(2),
    UInt32("crc32"),
    Pad(8),
    yelo_header,
    UEnum32('foot', ('foot', 'foot'), DEFAULT='foot', OFFSET=2044),
    SIZE=2048
    )

map_header_vap = desc_variant(
    map_header,
    ("yelo_header", Struct("vap_header", INCLUDE=vap_header, OFFSET=128)),
    )

tag_header = Struct("tag header",
    UEnum32("class 1", GUI_NAME="primary tag class", INCLUDE=valid_tags_os),
    UEnum32("class 2", GUI_NAME="secondary tag class", INCLUDE=valid_tags_os),
    UEnum32("class 3", GUI_NAME="tertiary tag class", INCLUDE=valid_tags_os),
    UInt32("id"),
    UInt32("path offset"),
    UInt32("meta offset"),
    UInt8("indexed"),
    Pad(3),
    # if indexed is non-zero, the meta_offset is the literal index in
    # the bitmaps, sounds, or loc cache that the meta data is located in.
    # NOTE: indexed is NOT a bitfield, if it is non-zero it is True
    UInt32("pad"),
    STEPTREE=CStrTagRef("path", POINTER=tag_path_pointer, MAX=768),
    SIZE=32
    )

tag_index_array = TagIndex("tag index",
    SIZE=".tag_count", SUB_STRUCT=tag_header, POINTER=tag_index_array_pointer
    )

tag_index_xbox = Struct("tag index",
    UInt32("tag index offset"),
    UInt32("scenario tag id"),
    UInt32("map id"),  # normally unused, but can be used
                       # for spoofing the maps checksum.
    UInt32("tag count"),

    UInt32("vertex parts count"),
    UInt32("model data offset"),

    UInt32("index parts count"),
    UInt32("index parts offset"),
    UInt32("tag sig", EDITABLE=False, DEFAULT='tags'),

    SIZE=36,
    STEPTREE=tag_index_array
    )

tag_index_pc = Struct("tag index",
    UInt32("tag index offset"),
    UInt32("scenario tag id"),
    UInt32("map id"),  # normally unused, but can be used
                       # for spoofing the maps checksum.
    UInt32("tag count"),

    UInt32("vertex parts count"),
    UInt32("model data offset"),

    UInt32("index parts count"),
    UInt32("vertex data size"),
    UInt32("model data size"),
    UInt32("tag sig", EDITABLE=False, DEFAULT='tags'),

    SIZE=40,
    STEPTREE=tag_index_array
    )

#tag_index_pc = tipc = dict(tag_index_xbox)
#tipc['ENTRIES'] += 1; tipc['SIZE'] += 4
#tipc[7] = UInt32("vertex data size")
#tipc[9] = tipc[8]; tipc[8] = UInt32("model data size")

map_header_def = BlockDef(map_header)
map_header_anni_def = BlockDef(map_header, endian=">")
map_header_demo_def = BlockDef(map_header_demo)
map_header_vap_def = BlockDef(map_header_vap)

tag_index_xbox_def = BlockDef(tag_index_xbox)
tag_index_pc_def = BlockDef(tag_index_pc)
tag_index_anni_def = BlockDef(tag_index_pc, endian=">")

map_header_vap_def = BlockDef(map_header_vap)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
THIS DEFINITION IS INCORRECT BECAUSE THE UNIT STRUCTURE IS DIFFERENT THAN HALO'S
'''

from ...hek.defs.unit import *
from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

unit_attrs = Struct("unit_attrs",
    Bool32("flags",
        "circular_aiming",
        "destroyed_after_dying",
        "half_speed_interpolation",
        "fires_from_camera",
        "entrance_inside_bounding_sphere",
        "unused",
        "causes_passenger_dialogue",
        "resists_pings",
        "melee_attack_is_fatal",
        "dont_reface_during_pings",
        "has_no_aiming",
        "simple_creature",
        "impact_melee_attaches_to_unit",
        "cannot_open_doors_automatically",
        "melee_attackers_cannot_attach",
        "not_instantly_killed_by_melee",
        "shield_sapping",
        "runs_around_flaming",
        "inconsequential",
        "special_cinematic_unit",
        "ignored_by_autoaiming",
        "shields_fry_infection_forms",
        "integrated_light_controls_weapon",
        "integrated_light_lasts_forever",
        "unknown24",
        "unknown25"
        ),
    ascii_str32('unknown1'),
    SEnum16('default_team', *unit_teams),
    SEnum16('constant_sound_volume', *sound_volumes),
    float_zero_to_inf('rider_damage_fraction'),
    dependency_stubbs('integrated_light_toggle', "effe"),
    SEnum16('A_in', *unit_inputs),
    SEnum16('B_in', *unit_inputs),
    SEnum16('C_in', *unit_inputs),
    SEnum16('D_in', *unit_inputs),
    float_rad('camera_field_of_view'),  # radians
    Float('camera_stiffness'),
    ascii_str32('camera_marker_name'),
    ascii_str32('camera_submerged_marker_name'),
    float_rad('pitch_auto_level'),  # radians
    from_to_rad('pitch_range'),  # radians
    reflexive("camera_tracks", camera_track, 2,
              'loose', 'tight'),

    BytearrayRaw('unknown2', SIZE=68),

    #Miscellaneous
    QStruct("seat_acceleration_scale", INCLUDE=ijk_float),
    Pad(12),
    float_zero_to_one('soft_ping_threshold'),  # [0,1]
    float_sec('soft_ping_interrupt_time', UNIT_SCALE=sec_unit_scale),  # seconds
    float_zero_to_one('hard_ping_threshold'),  # [0,1]
    float_sec('hard_ping_interrupt_time', UNIT_SCALE=sec_unit_scale),  # seconds
    float_zero_to_one('hard_death_threshold'),  # [0,1]
    float_zero_to_one('feign_death_threshold'),  # [0,1]
    float_sec('feign_death_time', UNIT_SCALE=sec_unit_scale),  # seconds
    float_wu('distance_of_evade_aim'),  # world units
    float_wu('distance_of_dive_aim'),  # world units

    Pad(4),
    float_zero_to_one('stunned_movement_threshold'),  # [0,1]
    float_zero_to_one('feign_death_chance'),  # [0,1]
    float_zero_to_one('feign_repeat_chance'),  # [0,1]
    dependency_stubbs('spawned_actor', "actv"),
    QStruct("spawned_actor_count",
        SInt16("from", GUI_NAME=""), SInt16("to"), ORIENT='h',
        ),
    float_wu_sec('spawned_velocity'),
    float_rad_sec('aiming_velocity_maximum',
                  UNIT_SCALE=irad_per_sec_unit_scale),  # radians/sec
    float_rad_sec_sq('aiming_acceleration_maximum',
                     UNIT_SCALE=irad_per_sec_sq_unit_scale),  # radians/sec^2
    float_zero_to_one('casual_aiming_modifier'),
    float_rad_sec('looking_velocity_maximum',
                  UNIT_SCALE=irad_per_sec_unit_scale),  # radians/sec
    float_rad_sec_sq('looking_acceleration_maximum',
                     UNIT_SCALE=irad_per_sec_sq_unit_scale),  # radians/sec^2

    Pad(8),
    Float('ai_vehicle_radius'),
    Float('ai_danger_radius'),
    dependency_stubbs('melee_damage', "jpt!"),
    SEnum16('motion_sensor_blip_size',
        "medium",
        "small",
        "large",
        ),
    Pad(2),

    Pad(12),  # open sauce unit extension padding
    reflexive("new_hud_interfaces", new_hud_interface, 2,
        'default/solo', 'multiplayer'),
    reflexive("dialogue_variants", dialogue_variant, 16,
        DYN_NAME_PATH='.dialogue.filepath'),

    #Grenades
    float_wu_sec('grenade_velocity'),
    SEnum16('grenade_type', *grenade_types),
    SInt16('grenade_count', MIN=0),

    Pad(4),
    reflexive("powered_seats", powered_seat, 2,
              "driver", "gunner"),
    reflexive("weapons", weapon, 4, DYN_NAME_PATH='.weapon.filepath'),
    reflexive("seats", seat, 16, DYN_NAME_PATH='.label'),

    SIZE=372
    )

unit_body = Struct('tagdata',
    unit_attrs,
    SIZE=372
    )

#def get():
#    return unit_def
del get

unit_def = TagDef("unit",
    blam_header_stubbs('unit', 2),
    unit_body,

    ext=".unit", endian=">"
    )



import re
import zlib

from math import log

from supyr_struct.util import *
from reclaimer.util import compression
from reclaimer.util import geometry
from reclaimer.util import matrices


POS_INF = float("inf")
NEG_INF = float("-inf")
RESERVED_WINDOWS_FILENAME_MAP = {}
INVALID_PATH_CHARS = set([str(i.to_bytes(1, 'little'), 'latin-1')
                          for i in (tuple(range(32)) +
                                    tuple(range(128, 256)))]
                         )
VALID_NUMERIC_CHARS = frozenset("0123456789")
for name in ('CON', 'PRN', 'AUX', 'NUL'):
    RESERVED_WINDOWS_FILENAME_MAP[name] = '_' + name
for i in range(1, 9):
    RESERVED_WINDOWS_FILENAME_MAP['COM%s' % i] = '_COM%s' % i
    RESERVED_WINDOWS_FILENAME_MAP['LPT%s' % i] = '_LPT%s' % i
INVALID_PATH_CHARS.update('<>:"|?*')


def is_reserved_tag(tag_index_ref):
    return (tag_index_ref.id == 0xFFffFFff and
            tag_index_ref.class_1.data == 0xFFFFFFFF)


def is_protected_tag(tagpath):
    return not tagpath or tagpath in RESERVED_WINDOWS_FILENAME_MAP or (
        not INVALID_PATH_CHARS.isdisjoint(set(tagpath)))


def is_overlapping_ranges(range_0, range_1):
    assert isinstance(range_0, range)
    assert isinstance(range_1, range)
    assert range_0.step == 1
    assert range_1.step == 1
    return ((range_0.start in range_1 or range_0.stop in range_1) or
            (range_1.start in range_0 or range_1.stop in range_0))


def get_is_xbox_map(engine):
    return "xbox" in engine or engine in ("stubbs", "shadowrun_proto")


def float_to_str(f, max_sig_figs=7):
    if f == POS_INF:
        return "1000000000000000000000000000000000000000"
    elif f == NEG_INF:
        return "-1000000000000000000000000000000000000000"

    sig_figs = -1
    if abs(f) > 0:
        sig_figs = int(round(max_sig_figs - log(abs(f), 10) - 1))

    if sig_figs < 0:
        return str(f).split(".")[0]

    str_float = ("%" + (".%sf" % sig_figs)) % f
    if "." in str_float:
        return str_float.rstrip("0").rstrip(".")
    return str_float


def float_to_str_truncate(f, sig_figs=7):
    if f == POS_INF:
        str_float = "1000000000000000000000000000000000000000"
    elif f == NEG_INF:
        str_float = "-1000000000000000000000000000000000000000"
    else:
        str_float = ("%" + (".%sf" % sig_figs)) % f

    float_pieces = str_float.split(".", 1)
    if len(float_pieces) == 1:
        remainder = "0" * sig_figs
    else:
        remainder = float_pieces[1] + "0" * (sig_figs - len(float_pieces[1]))

    return "%s.%s" % (float_pieces[0], remainder[: sig_figs])

# Test showing off the regex can be seen here:
# https://regex101.com/r/Fpd23n/1
JM_INT_PARSE_REGEX = re.compile(r'^\s*([-+]?\d+)')
# https://regex101.com/r/rKvo0t/1
# TODO: Should this match numbers that start with a . and are just commas?
# Check atof description.
JM_FLOAT_PARSE_REGEX = re.compile(r'^\s*([-+]?\d+\.?\d*)')

def parse_jm_int(string):
    '''
    convert jm integer to a proper integer. Based on description of C atoi spec.
    Returns 0 when it can't find anything or is interupted.
    '''
    result = JM_INT_PARSE_REGEX.search(string)
    return int(result.group()) if result else 0

def parse_jm_float(string):
    '''
    convert jm float to a proper float. Based on description of C atof spec.
    Returns 0.0 for NaN or when it can't find anything or is interupted.
    '''
    result = JM_FLOAT_PARSE_REGEX.search(string)
    return float(result.group()) if result else 0.0


def is_valid_ascii_name_str(string):
    if not string:
        return True

    try:
        string_bytes = set(string.encode("latin-1"))
    except Exception:
        return False

    if max(string_bytes) > 127:
        return False

    for i in tuple(range(8)) + tuple(range(14, 32)):
        if i in string_bytes:
            return False
    return True

def calc_halo_crc32(buffer, offset=None, size=None, crc=0xFFffFFff):
    if offset is not None:
        buffer.seek(offset)

    return zlib.crc32(buffer.read(size), crc ^ 0xFFffFFff) ^ 0xFFffFFff

del re



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Oh the xex has names.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


perf_performance = Struct("performance",
    Bool32("flags",
        ("disable_self_shadowing", 1 << 1),
        "disable_player_shadows",
        ),
    Float("water"),
    Float("decorators"),
    Float("effects"),
    Float("instanced_geometry"),
    Float("object_fade"),
    Float("object_lod"),
    Float("decals"),
    SInt32("cpu_light_count"),
    Float("cpu_light_quality"),
    SInt32("gpu_light_count"),
    Float("gpu_light_quality"),
    SInt32("shadow_count"),
    Float("shadow_quality"),
    ENDIAN=">", SIZE=56
    )


perf_body = Struct("tagdata",
    h3_reflexive("performance", perf_performance),
    ENDIAN=">", SIZE=12
    )


def get():
    return perf_def

perf_def = TagDef("perf",
    h3_blam_header('perf'),
    perf_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["perf"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get(): return cont_def

point_state = Struct("point_state",
    from_to_sec("state_duration"),
    from_to_sec("state_transition_duration"),
    dependency("physics", "pphy"),
    Pad(32),
    float_wu("width"),
    QStruct("color_lower_bound", INCLUDE=argb_float),
    QStruct("color_upper_bound", INCLUDE=argb_float),
    Bool32("scale_flags",
        "duration",
        "duration_delta",
        "transition_duration",
        "transition_duration_delta",
        "color",
        ),
    SIZE=104
    )

cont_body = Struct("tagdata",
    Bool16("flags",
        "first_point_unfaded",
        "last_point_unfaded",
        "points_start_pinned_to_media",
        "points_start_pinned_to_ground",
        "points_always_pinned_to_media",
        "points_always_pinned_to_ground",
        "edge_effect_fades_slowly",
        ),
    Bool16("scale_flags",
        "point_generation_rate",
        "point_velocity",
        "point_velocity_delta",
        "point_velocity_cone_angle",
        "inherited_velocity_fraction",
        "sequence_animation_rate",
        "texture_scale_u",
        "texture_scale_v",
        "texture_animation_u",
        "texture_animation_v",
        ),

    Struct("point_creation",
        Float("generation_rate",
               SIDETIP="points/sec", UNIT_SCALE=per_sec_unit_scale),
        from_to_wu_sec("velocity"),
        float_rad("velocity_cone_angle"),
        Float("inherited_velocity_fraction"),
        ),

    Struct("rendering",
        SEnum16("render_type",
            "vertical_orientation",
            "horizontal_orientation",
            "media_mapped",
            "ground_mapped",
            "viewer_facing",
            "double_marker_linked",
            ),
        Pad(2),
        Float("texture_repeats_u"),
        Float("texture_repeats_v"),

        Float("texture_animation_u", SIDETIP="repeats/sec"),
        Float("texture_animation_v", SIDETIP="repeats/sec"),
        Float("animation_rate", SIDETIP="frames/sec"),
        dependency("bitmap", "bitm"),
        SInt16("first_sequence_index"),
        SInt16("sequence_count"),
        Pad(100),

        FlUInt32("unknown0", VISIBLE=False),
        Bool16("shader_flags", *shader_flags),
        SEnum16("framebuffer_blend_function", *framebuffer_blend_functions),
        SEnum16("framebuffer_fade_mode", *render_fade_mode),
        Bool16("map_flags",
            "unfiltered",
            ),
        ),
    Pad(12),  # OS v4 shader extension padding
    Pad(16),

    Struct("secondary_map",
        dependency("bitmap", "bitm"),
        SEnum16("anchor", *render_anchor),
        Bool16("map_flags",
            "unfiltered",
            ),

        Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
        Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
        Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),
        QStruct("rotation_center", INCLUDE=xy_float),
        Pad(4),
        Float("zsprite_radius_scale"),
        Pad(20),
        ),
    reflexive("point_states", point_state, 16),
    SIZE=324,
    )

cont_def = TagDef("cont",
    blam_header('cont', 3),
    cont_body,

    ext=".contrail", endian=">", tag_cls=HekTag
    )
